__author__ = "Andrea Tramacere"


import numpy as np

from jetset.poly_fit import do_log_Parab_FIT
from jetset.loglog_poly_model import LogLinear
from jetset.model_manager import FitModel
from jetset.minimizer import fit_SED
from jetset.data_loader import Data,ObsData


def get_SED_pl_fit(j, comp, fit_range):
    """
    power-law fit of a jet SED spectral component
    Parameters
    ----------
    j: jet object
    comp: spectral component name
    fit_range: [nu1,nu2] range (log Hz)

    Returns
    -------
    best-fit-par
    best-fit-err
    pl-func

    """

    c = j.get_spectral_component_by_name(comp)
    x = c.SED.nu.value
    y = c.SED.nuFnu.value

    data = Data(n_rows=x.shape[0])
    data.set_field('x', x)
    data.set_field('y', y)
    # data.set_field('dy',value=1E-15)
    data.set_meta_data('z', j.parameters.z_cosm.val)
    data.set_meta_data('restframe', 'obs')
    data.set_meta_data('data_scale', 'lin-lin')

    sed_data = ObsData(data_table=data)
    loglog_poly = LogLinear()
    loglog_pl = FitModel(name=None, loglog_poly=loglog_poly)
    mm, best_fit = fit_SED(loglog_pl,
                           sed_data,
                           10 ** fit_range[0],
                           10 ** fit_range[1],
                           loglog=True,
                           silent=True,
                           fitname='spectral-indices-best-fit',
                           minimizer='lsb')

    par = loglog_pl.get_par_by_name(loglog_poly, 'alpha')

    return par.best_fit_val, par.best_fit_err, loglog_pl


def get_SED_log_par_fit(x_p,y_p,j,comp,delta_p=[-1,1]):
    """
    log-parabolic fit of a  jet SED spectral component

    Parameters
    ----------
    x_p: starting value (log)
    y_p: starting value (log)
    j: jet object
    comp
    delta_p: +/- boundaries centered on x_p fit range (log of nu in Hz), i.e. fit rage= [log10(nu_p) + delta_p[0], nu_p + log10(delta_p[1]) ]

    Returns
    -------
    p [x_p, y_p, curvature]
    err [err x_p, err y_p, err curvature]

    """
    c=j.get_spectral_component_by_name(comp)
    msk = c.SED.nuFnu.value > 0
    x = np.log10(c.SED.nu.value[msk])
    y = np.log10(c.SED.nuFnu.value[msk])
    #x=np.log10(c.SED.nu.value)
    #y=np.log10(c.SED.nuFnu.value)

    p,err=do_log_Parab_FIT(x,y,x_p,y_p,-0.1,x_range=[x_p + delta_p[0], x_p + delta_p[1] ],dy=np.ones(x.size))
    p,err=do_log_Parab_FIT(x,y,p[0],p[1],p[2],x_range=[p[0] + delta_p[0], p[0] + delta_p[1] ],dy=np.ones(x.size))

    return p, err


def get_nu_p_S_delta_approx(my_jet,gp):
    """
    estimates nu peak of synchrotron delta-approx

    Parameters
    ----------
    my_jet: jet object
    gp: gamma-peak

    Returns
    -------
    nu_p

    """
    B=my_jet.parameters.B.val
    delta=my_jet.get_beaming()
    z=my_jet.parameters.z_cosm.val
    return np.log10(3.2E6*B*delta/(1+z))+2*gp


def get_n_gamma_log_par_fit(n, power=0, delta_p=[-0.5, 0.5]):
    """
    estimates the peak of n(gamma)*gamma^3 by log-par fit
    Parameters
    ----------
    n: emitters distribution  object
    power: 0 for n(gamma), 1 for n(gamma)*gamma, 2 for n(gamma)*gamma^2 etc...
    delta_p: +/- boundaries centered on x_p fit range (log), i.e. fit rage= [log10(gamma_p) + delta_p[0], log10(gamma_p) + log10(delta_p[1]) ]


    Returns
    -------
    p [x_p, y_p, curvature]
    err [err x_p, err y_p, err curvature]
    """

    n.update()

    if n.emitters_type == 'electrons' :
        x = np.log10(n.gamma_e)
        y = np.log10(n.n_gamma_e)
    elif n.emitters_type == 'protons':
        x = np.log10(n.gamma_p)
        y = np.log10(n.n_gamma_p)
    else:
        raise RuntimeError('emitters must be protons or electrons')

    y = y + power * x

    y_p = y.max()
    x_p = x[np.argmax(y)]
    p, err = do_log_Parab_FIT(x, y, x_p, y_p, -1, x_range=[x_p + delta_p[0], x_p + delta_p[1]], dy=np.ones(x.size))
    return p,err



import pytest
from astropy.table import Table
from .base_class import TestBase


class TestCompositeModel(TestBase):

    def integration_suite(self,plot=False):
        self._all(plot=plot)

    def test_composite_model_pars(self,plot=False):
        from jetset.model_manager import  FitModel

        from jetset.template_2Dmodel import EBLAbsorptionTemplate
        ebl_franceschini=EBLAbsorptionTemplate.from_name('Franceschini_2008')

        from jetset.jet_model import Jet
        jet=Jet()

        jet.set_gamma_grid_size(100)
        fit_model=FitModel(jet=jet, name='test')

        fit_model.add_component(ebl_franceschini)
        fit_model.link_par(par_name='z_cosm', from_model='Franceschini_2008', to_model='jet_leptonic')
        fit_model.composite_expr='(jet_leptonic)*Franceschini_2008'
        fit_model.freeze('jet_leptonic','gmin')
        fit_model.parameters._build_par_table()
        fit_model.save_model('fit_model.pkl')
    
        _fit_model=FitModel.load_model('fit_model.pkl')
        _fit_model.jet_leptonic.parameters.gmin.frozen=False
        _fit_model.parameters._build_par_table()
        _t=_fit_model.parameters._par_table
        _t.sort(_t.colnames)
        

        _fit_model_clone=fit_model.clone()
        _fit_model_clone.jet_leptonic.parameters.gmin.frozen=False
        _fit_model_clone.parameters._build_par_table()
        _t_clone=_fit_model_clone.parameters._par_table
        _t_clone.sort(_t.colnames)
    

        assert(all([_t[i]['frozen']==_t_clone[i]['frozen'] for i in range(len(_t))]))
    
        assert(all([_t[i]['val']==_t_clone[i]['val'] for i in range(len(_t)) if _t.mask[i] is False]))
        
        fit_model.jet_leptonic.parameters.gmin.frozen=False
        fit_model.parameters._build_par_table()
        t=fit_model.parameters._par_table
        t.sort(_t.colnames)
        
        
        assert(all([_t[i]['frozen']==t[i]['frozen'] for i in range(len(_t))]))
        assert(all([_t[i]['val']==t[i]['val'] for i in range(len(_t)) if _t.mask[i] is False]))
        
        for _m in [fit_model,_fit_model,_fit_model_clone]:
            p_m=_m.parameters.get_par_by_name('jet_leptonic','z_cosm')
            p_l=_m.parameters.get_par_by_name('Franceschini_2008','z_cosm')
            assert( p_m._is_dependent == False )
            assert( p_l._is_dependent == True )
            p_l_m=p_l._master_pars[0]
            assert(p_l_m.name == p_m.name)



__author__ = "Andrea Tramacere"


import numpy as np
import os
from astropy.table import Table
from numpy.core._multiarray_umath import zeros, log10
from scipy import interpolate

on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

if on_rtd == True:
    try:
        from .jetkernel import jetkernel as BlazarSED
    except ImportError:
        from .mock import jetkernel as BlazarSED
else:
    from .jetkernel import jetkernel as BlazarSED

from . import spectral_shapes
from .jetkernel_models_dic import nuFnu_obs_dict, n_seed_dic
from .plot_sedfit import PlotSpecComp,PlotSeedPhotons
from .utils import check_frame, unexpected_behaviour

__all__=['JetSeedPhotons','JetSpecComponent','SpecCompList']

class JetSeedPhotons(object):
    """

    """
    def __init__(self,name,blob_object,var_name=None):
        self.name = name

        self._blob_object = blob_object
        self._n_name, self._nu_name = n_seed_dic[self.name]

        self.n_ptr = getattr(blob_object, self._n_name)

        self.nu_ptr = getattr(blob_object, self._nu_name)
        #self.SED = spectral_shapes.SED(name=self.name)
        if var_name is not None:
            self._var_name=var_name

        self.fill(emiss_lim=self._blob_object.emiss_lim)

    def fill(self,log_log=False,emiss_lim=0):
        self.nu,self.n=self.get_spectral_points(log_log=log_log,emiss_lim=emiss_lim)

    def get_spectral_points(self,log_log=False,emiss_lim=0):

        #try:

        size=self._blob_object.nu_grid_size
        x=zeros(size)
        y=zeros(size)

        for i in range(size):
            x[i]=BlazarSED.get_spectral_array(self.nu_ptr,self._blob_object,i)
            y[i]=BlazarSED.get_spectral_array(self.n_ptr,self._blob_object,i)

            #print("->%e %e"%(x[i],y[i]))

        msk_nan=np.isnan(x)
        msk_nan+=np.isnan(y)
        #print('emiss lim',self.get_emiss_lim())
        x[msk_nan]=0.
        y[msk_nan]=emiss_lim

        msk=y<emiss_lim


        y[msk]=emiss_lim



        if log_log==True:
            msk = y <= 0.
            y[msk] = emiss_lim

            #x=x[msk]
            #    y=y[msk]

            x=log10(x)
            y=log10(y)



        return x,y

        #except:
        #    raise RuntimeError ('model evaluation failed in get_spectral_points')


    def plot(self, y_min=None,y_max=None):
        self.fill(emiss_lim=self._blob_object.emiss_lim)
        p=PlotSeedPhotons()
        p.plot(nu=self.nu,nuFnu=self.n,y_min=y_min,y_max=y_max)

        return p



class JetSpecComponent(object):
    """
    """

    def __repr__(self):
        return str(self.show())

    #def __str__(self):
    #    return str(self.show())


    def __init__(self,jet_obj,name,blob_object,var_name=None,state_dict=None,state=None):

        self.name=name
        self.jet_obj=jet_obj

        self._blob_object=blob_object
        self._nuFnu_name, self._nu_name=nuFnu_obs_dict[self.name]

        self.nuFnu_ptr=getattr(blob_object,self._nuFnu_name)

        self.nu_ptr=getattr(blob_object,self._nu_name)

        self.SED=spectral_shapes.SED(name=self.name,beaming=jet_obj.get_beaming())
        self.seed_field=None

        # self._nu_start_src_name, self._nu_stop_src_name = nu_src_start_stop_dict[self.name]
        #
        # self.nu_ptr_start = getattr(blob_object, self._nu_name)
        # self.nu_ptr_stop = getattr(blob_object, self._nu_name)
        #
        # self._nu_start_src = 'auto'
        # self._nu_stop_src = 'auto'
        # self._nu_start_obs = 'auto'
        # self._nu_stop_obs = 'auto'

        if name in n_seed_dic.keys():
            self.seed_field=JetSeedPhotons(name,blob_object)


        if var_name is not None:
            self._var_name=var_name

            if state_dict is None:
                self._state_dict = dict()
                self._state_dict['on'] = 1
                self._state_dict['off'] = 0
            else:
                self._state_dict=state_dict
            self.state='on'
        else:
            self._state_dict = {}
            self._var_name=None
            self._state='on'

        if state is not None and self._state_dict != {}:
            self.state=state

    # @property
    # def nu_boundaries(self,frame='obs'):
    #     check_frame(frame)
    #     if frame == 'src':
    #         return self._nu_start_src, self._nu_stop_src
    #     else:
    #         return self._nu_start_obs, self._nu_stop_obs
    #
    # @nu_boundaries.setter
    # def nu_boundaries(self,nu_start=None,nu_stop=None, frame='obs'):
    #     check_frame(frame)
    #     if frame == 'obs':
    #         self._nu_start_obs = nu
    #         self._nu_start_src = convert_nu_to_src(nu,self.jet_obj.get_par_by_type('redshift').val,'obs')
    #     else:
    #         self._nu_start_src = nu
    #         self._nu_start_obs = convert_nu_to_src(nu, self.jet_obj.get_par_by_type('redshift').val, 'obs')

    def get_emiss_lim(self,seed=False):
        return self._blob_object.emiss_lim


    def fill_SED(self,log_log=False,lin_nu=None,skip_zeros=False):

        x,y=self.get_SED_points( log_log=log_log,lin_nu=lin_nu,skip_zeros=skip_zeros)
        self.SED.beaming=self.jet_obj.get_beaming()
        self.SED.fill(nu=x,nuFnu=y,log_log=log_log)
        self.SED.fill_nuLnu(z=self.jet_obj.get_par_by_type('redshift').val,dl=self.jet_obj.get_DL_cm())

        if self.seed_field is not None:
            self.seed_field.fill(log_log=log_log)



    def get_SED_points(self, log_log=False, lin_nu=None,interp='linear',skip_zeros=False):

        size = self._blob_object.nu_grid_size
        x = zeros(size)
        y = zeros(size)

        for i in range(size):
            x[i] = BlazarSED.get_spectral_array(self.nu_ptr, self._blob_object, i)
            y[i] = BlazarSED.get_spectral_array(self.nuFnu_ptr, self._blob_object, i)


        msk_nan = np.isnan(x)
        msk_nan += np.isnan(y)

        x[msk_nan] = 0.
        y[msk_nan] = self.get_emiss_lim()

        msk = y < self.get_emiss_lim()
        y[msk] = self.get_emiss_lim()

        msk_zeros = y > self.get_emiss_lim()

        if lin_nu is not None:
            #f_interp=interpolate.Akima1DInterpolator(log10(x), log10(y))
            f_interp = interpolate.interp1d(log10(x), log10(y), bounds_error=False, kind=interp)
            y = np.power(10., f_interp(log10(lin_nu)))
            x=lin_nu
            msk_nan = np.isnan(y)
            y[msk_nan] = 0.
            msk_zeros = y > self.get_emiss_lim()
            y[~msk_zeros] = 0.

        if log_log == True:
            msk = y <= 0.
            y[msk] = -1.0E10
            msk_zeros = y >  self.get_emiss_lim()
            x = log10(x)
            y = log10(y)


        if skip_zeros is True:
            _x = x[msk_zeros]
            _y = y[msk_zeros]
        else:
            _x = x
            _y = y

        return _x, _y




    def update(self):
        size = self._blob.nu_grid_size
        x = zeros(size)
        y = zeros(size)

        for i in range(size):
            x[i] = BlazarSED.get_spectral_array(self.nu_ptr, self._blob, i)
            y[i] = BlazarSED.get_spectral_array(self.nuFnu_ptr, self._blob, i)


    def show(self):
        print('name                :',self.name)
        print('var name            :',self._var_name)
        print('state               :',self._state)
        #print('nu_start (src frame):', self._nu_start_src)
        #print('nu_stop  (src frame):', self._nu_stop_src)
        if self._state_dict is not None:
            print('allowed states :',[k for k in self._state_dict.keys()])


    @property
    def state(self,):
        return self._state

    @state.setter
    def state(self, val):
        if self._state_dict!={}:
            if val not in self._state_dict.keys():
                raise RuntimeError('val', val, 'not in allowed', self._state_dict.keys())
            self._state = val
            if self._var_name is not None:
                setattr(self._blob_object, self._var_name, self._state_dict[val])
        else:
            raise Warning('the state of the spectral component',self.name,' can not be changed')

    def get_var_state(self,):
        if self._var_name is not None:
            return  getattr(self._blob_object,self._var_name)
        else:
            return None



    def plot(self, y_min=None,y_max=None):
        p=PlotSpecComp()
        p.plot(nu=self.SED.nu.value,nuFnu=self.SED.nuFnu.value,y_min=y_min,y_max=y_max)

        return p


class SpecCompList(object):

    def __init__(self,sc_list):
        self._sc_list=sc_list
        self._table=None


    def show(self):
        for sc in self._sc_list:
            sc.show()

    def __repr__(self):
        return str(self.show())

    def build_table(self, restframe='obs'):

        _names = ['nu']
        _cols=[]

        check_frame(restframe)
        if restframe=='obs':
           _cols.append(self._sc_list[0].SED.nu)
        elif restframe=='src':
            _cols.append(self._sc_list[0].SED.nu_src)
        else:
            unexpected_behaviour()

        for ID,sc in enumerate(self._sc_list):
            _names.append(sc.name)
            if restframe == 'obs':
                _cols.append(sc.SED.nuFnu)
            else:
                _cols.append(sc.SED.nuLnu_src)



        _meta=dict(src_name=sc.jet_obj.name)
        _meta['redshift']=sc.jet_obj.get_par_by_type('redshift').val
        _meta['restframe']= restframe
        self._table = Table(_cols, names=_names,meta=_meta)

    def get_spectral_component_by_name(self,name,verbose=True):
        for i in range(len(self._sc_list)):
            if self._sc_list[i].name==name:
                return self._sc_list[i]
        else:
            if verbose==True:
                print ("no spectral components with name %s found"%name)


    @property
    def table(self):
        return self._table


__author__ = "Andrea Tramacere"

import os
import shutil



__author__ = "Andrea Tramacere"

__all__=['clean_dir','makedir','section_separator', 'WorkPlace']




def clean_dir(dir_name):
    print ("cleaning dir",dir_name)
    shutil.rmtree(dir_name)
    os.mkdir(dir_name)

def makedir(out_dir,clean_work_dir=True):
    """
    creates a directory
    """
    if os.path.isdir(out_dir):
        Warning ("directory %s already existing"%(out_dir))
        if clean_work_dir==True:
            Warning ('removing existing dir')
            shutil.rmtree(out_dir)
            os.mkdir(out_dir)
            Warning ('the directory %s has been created' % (out_dir))
    else:
        if os.path.isfile(out_dir):
            Warning ("a file with the same name of dir=%s, exists"%out_dir)
            Warning ("select a differn name")
        else:
            os.mkdir(out_dir)
            Warning ('the directory %s has been created'%(out_dir))


class WorkPlace(object):
    """
    Class to set the working place
    using static members (class members)
    
    Variables
    
    :ivar out_dir: directory name (default=./) 
    :ivar flag: flag name (default=sed-fit-tests)
    
    
    """
    def __init__(self,out_dir='./jet_wd',flag='sed-fit-tests',clean=False):
        self.out_dir=out_dir
        self.flag=flag


        makedir(out_dir,clean_work_dir=clean)

    
    
            
            

section_separator = "="*80 + "\n"




__author__ = "Andrea Tramacere"

from itertools import cycle


import scipy as sp

import numpy as np

import warnings

from scipy.stats import chi2

import dill as pickle

try:
    import iminuit
    from iminuit import Minuit
    minuit_installed=True
except:
    minuit_installed=False


from tqdm.auto import tqdm


from .plot_sedfit import plt

from scipy.optimize import least_squares

from .leastsqbound.leastsqbound import leastsqbound

from .output import section_separator,WorkPlace,makedir
from .utils import JetkerneltException

from .data_loader import ObsData,Data
from .data_loader import lin_to_log





__all__ = ['FitResults','fit_SED','Minimizer','LSMinimizer','LSBMinimizer', 'MinuitMinimizer', 'ModelMinimizer']



class FitResults(object):
    """
    Class to store the fit results 
    
    Parameters
     
    
    Members
    :ivar fit_par: fit_par
    :ivar info: info
    :ivar mesg: mesg
    :ivar success: success
    :ivar chisq: chisq
    :ivar dof: dof
    :ivar chisq_red: chisq_red
    :ivar null_hyp_sig: null_hyp_sig
    :ivar fit_report: ivar get_report()
    -------
    
    """
    
    def __init__(self,
                 name,
                 mm,
                 parameters,
                 calls,
                 mesg,
                 success,
                 chisq,
                 dof,
                 chisq_red,
                 null_hyp_sig,
                 wd,
                 chisq_no_UL=None,
                 dof_no_UL=None,
                 chisq_red_no_UL=None,
                 null_hyp_sig_no_UL=None):

        self.name=name
        self.parameters=parameters
        self.mm=mm
        self.calls=calls
        self.mesg=mesg
        self.success=success
        self.chisq=chisq
        self.dof=dof
        self.chisq_red=chisq_red
        self.null_hyp_sig=null_hyp_sig

        self.chisq_no_UL = chisq_no_UL
        self.dof_no_UL = dof_no_UL
        self.chisq_red_no_UL = chisq_red_no_UL
        self.null_hyp_sig_no_UL = null_hyp_sig_no_UL

        self.update_report()
        self.wd=wd
         

    def update_report(self):
        # self.model_table=self.parameters._par_table
        self.parameters._build_best_fit_par_table()
        self.parameters._build_par_table()

    def _show_report(self,):
        self.update_report()

        print('-------------------------------------------------------------------------')
        print("Fit report")
        print("")
        print("Model: %s"%self.name)

        self.parameters.show_pars()
        print("")
        print("converged=%s"%self.success)
        print("calls=%d"%self.calls)
        print("mesg=")
        if hasattr(self,'mesg'):
            try:
                from IPython.display import display
                display(self.mesg)
            except:
                print(self.mesg)

        print("dof=%d"%self.dof)
        print("chisq=%f, chisq/red=%f null hypothesis sig=%f"%(self.chisq,self.chisq_red,self.null_hyp_sig))
        if self.dof_no_UL is not None:
            print("")
            print("stats without the UL")
            print("dof  UL=%d" % self.dof_no_UL)
            print(
                "chisq=%f, chisq/red=%f null hypothesis sig=%f" % (self.chisq_no_UL, self.chisq_red_no_UL, self.null_hyp_sig_no_UL))
            print("")
        print("")
        print("best fit pars")
        self.parameters.show_best_fit_pars()

        print('-------------------------------------------------------------------------')
        print("")

    @property
    def bestfit_table(self):
        return self.parameters.best_fit_par_table

    def _update_asymm_errors(self):
        if hasattr(self.mm):
            for pi in range(len(self.mm.fit_par_free)):
                if  hasattr(self.mm.minimizer,'asymm_errors'):
                    self.mm.fit_par_free[pi].err_p=self.mm.minimizer.asymm_errors[pi][0]
                    self.mm.fit_par_free[pi].err_m=self.mm.minimizer.asymm_errors[pi][1]

    def show_report(self):
        try:
            self._show_report()
        except:
            print('problem in formatting text for report')
        #except Exception as e:
        #    raise(RuntimeWarning,'problem in formatting text for report',e)

    @classmethod
    def load_report(cls,file_name):
        c = pickle.load(open(file_name, "rb"))
        return c

    def save_report(self,name=None):
        if name is None:
            name = 'best_fit_report.pkl'
        mm =self.mm
        self.mm=None
        setattr(self, 'mm', mm)
        try:
            pickle.dump(self, open(name, 'wb'), protocol=pickle.HIGHEST_PROTOCOL)
        except:
            o = FitResults(name=self.name,
                           mm=None,
                           parameters=self.parameters,
                           calls=self.calls,
                           mesg=None,
                           success=self.success,
                           chisq=self.chisq,
                           dof=self.dof,
                           chisq_red=self.chisq_red,
                           null_hyp_sig=self.null_hyp_sig,
                           wd=self.wd,
                           chisq_no_UL=self.chisq_no_UL,
                           dof_no_UL=self.dof_no_UL,
                           chisq_red_no_UL=self.chisq_red_no_UL,
                           null_hyp_sig_no_UL=self.null_hyp_sig_no_UL)

            pickle.dump(o, open(name, 'wb'), protocol=pickle.HIGHEST_PROTOCOL)












class ModelMinimizer(object):


    def __init__(self,minimizer_type):
        __accepted__ = ['lsb', 'minuit', 'ls','sherpa']


        if minimizer_type == 'lsb':
            self.minimizer=LSBMinimizer(self)

        elif minimizer_type=='ls':
            self.minimizer=LSMinimizer(self)

        elif minimizer_type=='minuit':
            self.minimizer=MinuitMinimizer(self)

        elif minimizer_type == 'sherpa':
            from .sherpa_plugin import  SherpaMinimizer
            self.minimizer = SherpaMinimizer(self)

        elif minimizer_type not in __accepted__:
            raise RuntimeError('minimizer ', minimizer_type, 'not accepted, please choose among', __accepted__)

        else:
            raise RuntimeError('minimizer factory failed')

        self.minimizer_type=minimizer_type
        #print('minimizer',minimizer_type)

    def save_model(self, file_name):
        _m= self.minimizer
        self.minimizer=self.minimizer_type

        pickle.dump(self, open(file_name, 'wb'), protocol=pickle.HIGHEST_PROTOCOL)
        self.minimizer =_m

    @classmethod
    def load_model(cls, file_name):

        try:
            c = pickle.load(open(file_name, "rb"))
            if isinstance(c, ModelMinimizer):
                c.__init__(c.minimizer)
                return  c
            else:
                raise RuntimeError('The model you loaded is not valid please check the file name')

        except Exception as e:
            raise RuntimeError(e)

    def _handle_sed_data(self,sed_data,loglog=False,use_fake_err=False):
        if sed_data.data['dnuFnu_data'] is None:
            sed_data.data['dnuFnu_data'] = np.ones(sed_data.data['nu_data'].size)

        if loglog is False:
            nu_fit = sed_data.data['nu_data']
            nuFnu_fit = sed_data.data['nuFnu_data']
            if use_fake_err == False:
                err_nuFnu_fit = sed_data.data['dnuFnu_data']
            else:
                err_nuFnu_fit = sed_data.data['dnuFnu_fake']
        else:
            nu_fit = sed_data.data['nu_data_log']
            nuFnu_fit = sed_data.data['nuFnu_data_log']
            err_nuFnu_fit = sed_data.data['dnuFnu_data_log']
            if use_fake_err == False:
                err_nuFnu_fit = sed_data.data['dnuFnu_data_log']
            else:
                err_nuFnu_fit = sed_data.data['dnuFnu_fake_log']

        UL = sed_data.data['UL']

        data = np.zeros(nu_fit.size,
                             dtype=[('x', nu_fit.dtype), ('y', nuFnu_fit.dtype), ('dy', err_nuFnu_fit.dtype),
                                    ('UL', UL.dtype)])
        data['x'] = nu_fit
        data['y'] = nuFnu_fit
        data['dy'] = err_nuFnu_fit
        data['UL'] = UL
        return data

    def _handle_data_xy(self, data, loglog=False, use_fake_err=False):
        #if data['dy'] is None:
        #    data['dy'] = np.ones(data['x'].size)

        if loglog is False:
            x = data.table['x']
            y = data.table['y']
            if use_fake_err == False:
                dy = data.table['dy']
            else:
                dy = data.table['dy_fake']
        else:
            x = np.log10(data.table['x'])
            y, dy = lin_to_log(val=data.table['y'], err=data.table['dy'])
            if use_fake_err == False:
                pass
            else:
                y, dy = lin_to_log(val=data.table['y'], err=data.table['dy_fake'])

        UL = data.table['UL']

        _data = np.zeros(x.size,
                        dtype=[('x', x.dtype), ('y', y.dtype), ('dy', dy.dtype),
                               ('UL', UL.dtype)])
        _data['x'] = x
        _data['y'] = y
        _data['dy'] = dy
        _data['UL'] = UL
        return _data


    def _prepare_fit(self,
                     fit_model,
                     data,
                     nu_fit_start,
                     nu_fit_stop,
                     fitname=None,
                     fit_workplace=None,
                     loglog=False,
                     silent=False,
                     get_conf_int=False,
                     use_fake_err=False,
                     use_UL=False):
        #print('--> DEBUG WorkPlace.flag',WorkPlace.flag,fit_workplace)

        if fitname is None:
            fitname = fit_model.name
            #print('--> DEBUG ')

        if fit_workplace is None:
            fit_workplace = WorkPlace()
            out_dir = fit_workplace.out_dir + '/' + fitname + '/'
        else:
            out_dir = fit_workplace.out_dir + '/' + fitname + '/'

        makedir(out_dir)

        for model in fit_model.components.components_list:
            if model.model_type == 'jet':
                model.set_path(out_dir)

        if isinstance(data,ObsData):
            self.data=self._handle_sed_data(data,use_fake_err=use_fake_err,loglog=loglog)
        elif isinstance(data,Data):
            self.data= self._handle_data_xy(data,use_fake_err=use_fake_err,loglog=loglog)
        else:
            raise  RuntimeError('data type must be Obs  Data or Data')

        #if sed_data.data['dnuFnu_data'] is None:
        #    sed_data.data['dnuFnu_data'] = np.ones(sed_data.data['nu_data'].size)

        # filter data points
        if loglog is True:
            nu_fit_start = np.log10(nu_fit_start)
            nu_fit_stop = np.log10(nu_fit_stop)

        msk1 = self.data['x'] > nu_fit_start
        msk2 = self.data['x'] < nu_fit_stop
        msk_zero_error = self.data['dy'] > 0.0
        self.use_UL=use_UL
        if use_UL == True:
            msk = msk1 * msk2 * msk_zero_error

        else:
            msk = msk1 * msk2 * msk_zero_error*~self.data['UL']

        self.data = self.data[msk]

        if silent == False:
            print("filtering data in fit range = [%e,%e]" % (nu_fit_start, nu_fit_stop))
            print("data length", self.data.size)



        for par in fit_model.parameters.par_array:
            if get_conf_int == True:
                par.set_fit_initial_value(par.best_fit_val)
            else:
                par.set_fit_initial_value(par.val)

        fit_par_free = [par for par in fit_model.parameters.par_array if par.frozen is False and (par.linked is False or par.root is True) and par._is_dependent is False]

        #remove duplicates and keeps the order
        fit_par_free = sorted(set(fit_par_free), key=fit_par_free.index)

        pinit = [par.get_fit_initial_value() for par in fit_par_free]

        # bounds
        free_pars = 0

        for pi in range(len(fit_model.parameters.par_array)):

            if fit_model.parameters.par_array[pi].frozen == False:
                free_pars += 1

        if silent == False:
            print(section_separator)
            print("*** start fit process ***")
            #print("initial pars: ")
            #fit_model.parameters.show_pars()
            print("----- ")

        self.out_dir=out_dir
        self.pinit=pinit
        self.pout=None
        self.free_pars=free_pars
        self.fit_par_free=fit_par_free
        self.fit_model=fit_model
        self.loglog=loglog

        if hasattr(self.fit_model,'nu_min_fit'):
            self.fit_model.nu_min_fit = nu_fit_start
            self.fit_model.nu_max_fit = nu_fit_stop



        if len(fit_par_free)>self.data['x'].size and isinstance(self.minimizer,LSBMinimizer):
            m='number of data points: %d is lower than number of free pars: %d'%(self.data['x'], len(fit_par_free))
            raise  JetkerneltException(message=m)

    def fit(self,
            fit_model,
            sed_data,
            nu_fit_start,
            nu_fit_stop,
            fitname=None,
            fit_workplace=None,
            loglog=False,
            silent=False,
            get_conf_int=False,
            max_ev=0,
            use_fake_err=False,
            use_UL=False,
            skip_minimizer=False,
            repeat=1):

        self.silent=silent
        #print('-->nu_fit_start', nu_fit_start)

        self._prepare_fit( fit_model, sed_data, nu_fit_start, nu_fit_stop, fitname=fitname, fit_workplace=fit_workplace,
                     loglog=loglog, silent=silent, get_conf_int=get_conf_int, use_fake_err=use_fake_err,use_UL=use_UL)

        fit_model.set_nu_grid(nu_min=nu_fit_start*0.5, nu_max=nu_fit_stop*1.5)


        for i in range(repeat):
            if skip_minimizer == False:
                if repeat>1:
                    if silent is False:
                        print('fit run:',i)
                    if i>0:
                        self.pinit = [v for v in self.minimizer.pout]
                        if silent is False:
                            print('- old chisq=%5.5e' % (self.minimizer.chisq))
                        old_chisq = self.minimizer.chisq

                self.minimizer.fit(self,max_ev=max_ev,silent=silent,use_UL=use_UL)

                self.pout = self.minimizer.pout
                self.errors = self.minimizer.errors
                if hasattr(self.minimizer, 'asymm_errors'):
                    self.asymm_errors = self.minimizer.asymm_errors


                self.reset_to_best_fit()
                self.minimizer._fit_stats()
                if silent is False:
                    print('- best chisq=%5.5e'%(self.minimizer.chisq))
                    print()

                new_chisq = self.minimizer.chisq
                if i==0:
                    old_chisq = new_chisq

                chisq_min = min(new_chisq, old_chisq)
                #if i==0 or new_chisq<=chisq_min:
                #    self.pout = self.minimizer.pout
                #    self.errors = self.minimizer.errors
                #    if  hasattr(self.minimizer,'asymm_errors'):
                #        self.asymm_errors = self.minimizer.asymm_errors
                #elif new_chisq>=chisq_min:
                #    self.pout = self.minimizer.pout
                #    self.errors = self.minimizer.errors
                #    if hasattr(self.minimizer, 'asymm_errors'):
                #        self.asymm_errors = self.minimizer.asymm_errors

                #if i<repeat-1:
                #    self.pinit = [par.val for par in self.fit_par_free]
                #    if silent is False:
                #        print()
            else:
                pass


        return self.get_fit_results(fit_model,nu_fit_start,nu_fit_stop,fitname,loglog=loglog,silent=silent)

    def get_fit_results(self, fit_model, nu_fit_start, nu_fit_stop, fitname, silent=False, loglog=False):

        self.reset_to_best_fit()
        self.minimizer._fit_stats()
        best_fit = FitResults(fitname,
                              self,
                              fit_model.parameters,
                              self.minimizer.calls,
                              self.minimizer.mesg,
                              self.minimizer.success,
                              self.minimizer.chisq,
                              self.minimizer.dof,
                              self.minimizer.chisq_red,
                              self.minimizer.null_hyp_sig,
                              self.out_dir,
                              chisq_no_UL=self.minimizer.chisq_no_UL,
                              dof_no_UL=self.minimizer.dof_no_UL,
                              chisq_red_no_UL=self.minimizer.chisq_red_no_UL,
                              null_hyp_sig_no_UL=self.minimizer.null_hyp_sig_no_UL)

        if silent == False:
            best_fit.show_report()

        fit_model.set_nu_grid(nu_min=nu_fit_start , nu_max=nu_fit_stop)
        fit_model.eval(fill_SED=True, loglog=loglog, phys_output=True)

        res_bestfit = self.minimizer.residuals_Fit(self.minimizer.pout,
                                                 self.fit_par_free,
                                                 self.data,
                                                 self.fit_model,
                                                 self.loglog,
                                                 use_UL=self.use_UL)



        if loglog == True:
            fit_model.SED.fill(nu_residuals=np.power(10, self.data['x']), residuals=res_bestfit)
        else:
            fit_model.SED.fill(nu_residuals=self.data['x'], residuals=res_bestfit)


        if silent == False:
            print(section_separator)

        fit_model.eval(fill_SED=True)
        if self.minimizer._post_fit_warnings!='':
                print('there are  fit warnings messages, use  the .show_fit_warnings() or access the member .minimizer._post_fit_warnings')

        return best_fit

    def show_fit_warnings(self):
        print(self.minimizer._post_fit_warnings)


    def reset_to_best_fit(self):

        for pi in range(len(self.fit_par_free)):
            self.fit_par_free[pi].set(val=self.pout[pi])
            self.fit_par_free[pi].best_fit_val = self.pout[pi]
            self.fit_par_free[pi].best_fit_err = self.errors[pi]
            if  hasattr(self,'asymm_errors'):
                self.fit_par_free[pi].err_p=self.asymm_errors[pi][0]
                self.fit_par_free[pi].err_m=self.asymm_errors[pi][1]

        self.fit_model.eval()


class Minimizer(object):

    def __init__(self,model):
        self.model=model
        self._progress_iter = cycle(['|', '/', '-', '\\'])
        self._post_fit_warnings=''
        self.pbar=None
            #self._progess_bar(pbar, _res_sum, _res_sum_UL)

    def fit(self,model,
            max_ev=None,
            use_UL=False,
            silent=False):
        if silent is False:
            self.pbar = tqdm(total=None)
        self.use_UL = use_UL
        self.calls=0
        self.res_check=None
        self.model=model
        self.silent=silent
        self._original_silent_state = silent
        self._fit(max_ev)
        self._fit_stats()
        self._set_fit_errors()


    def _force_silent(self):
        self.silent=True

    def _restore_silent_state(self):
        self.silent = self._original_silent_state


    def _fit_stats(self):
        self.dof = len(self.model.data['x']) - self.model.free_pars
        self.chisq = self.get_chisq()
        self.chisq_red = self.chisq / float(self.dof)
        self.null_hyp_sig = 1.0 - chi2.cdf(self.chisq, self.dof)
        if self.use_UL==True:
            self.dof_no_UL = len(self.model.data['x']) - self.model.free_pars - self.model.data['UL'].sum()
            self.use_UL=False
            self.chisq_no_UL = self.get_chisq()
            self.use_UL = True
            self.chisq_red_no_UL = self.chisq_no_UL / float(self.dof_no_UL)
            self.null_hyp_sig_no_UL=1.0 - chi2.cdf(self.chisq_no_UL, self.dof_no_UL)
        else:
            self.dof_no_UL = None
            self.chisq_red_no_UL = None
            self.chisq_no_UL=None
            self.null_hyp_sig_no_UL=None
        self.success = True
        self.status = 1



    def _set_fit_errors(self):
        if self.covar is None:
            print("!Warning, no covariance matrix produced")
            self.errors = np.zeros(len(self.pout))
        else:
            self.errors = [np.sqrt(np.fabs(self.covar[pi, pi]) * self.chisq_red) for pi in range(len(self.model.fit_par_free))]


    def _progess_bar(self, _res_sum, res_sum_UL):
        #if self.calls==1:
        #    print("minim function calls=%d, chisq=%5.5e UL part=%f" %(self.calls, _res_sum, -2.0*np.sum(res_UL)))
        #    print()
        if (np.mod(self.calls, 10) == 0 and self.calls != 0)  :
            #_c= ' ' * 256
            #print("\r%s"%_c,end="")
            #print("\r%s minim function calls=%d, chisq=%5.5e UL part=%f" % (next(self._progress_iter),self.calls, _res_sum, res_sum_UL), end="")
            m="minim. function calls=%d, chisq=%5.5e UL part=%f" %(self.calls, _res_sum, res_sum_UL)
            self.pbar.n=self.calls
            self.pbar.set_description(m)
            self.pbar.update(1)

    def residuals_Fit(self,
                      p,
                      fit_par,
                      data,
                      best_fit_SEDModel,
                      loglog,
                      chisq=False,
                      use_UL=False,
                      silent=False):
        #print('==>p', p)
        #print('==>fit_par_fre', self.model.fit_par_free)
        _warn=False
        for pi in range(len(fit_par)):
            _old_v=fit_par[pi].val
            #print('==> set', fit_par[pi].val,fit_par[pi].val_min,p[pi])
            fit_par[pi].set(val=p[pi])
            if np.isnan(p[pi]):
                _warn=True
                s1 = str('warning nan for par %s' % fit_par[pi].name)
                s2 = ', old parameter value was %s' % str(fit_par[pi].val)
                s =s1 + s2
                warnings.warn(s)
                self._post_fit_warnings+=s+'\n'
        #if _warn==True:
        #    best_fit_SEDModel.show_pars()
        #    print('res_sum',self._res_sum_chekc)
        #    print('res_chekc', self._res_chekc)
        #    print('res_UL_chekc', self._res_UL_chekc)


        model = best_fit_SEDModel.eval(nu=data['x'], fill_SED=False, get_model=True, loglog=loglog)


        _res_sum, _res, _res_sum_UL=_eval_res(data['y'], model, data['dy'], data['UL'], use_UL=use_UL)

        self._res_sum_chekc=_res_sum
        self._res_chekc = _res
        self._res_UL_chekc = _res_sum_UL
        #self._par_check=p
        #print('--> model', model[0],_res_sum)
        self.calls +=1

        if silent==False:
            #with tqdm(total=None, file=sys.stdout) as pbar:
            if self.pbar is not None:
                self._progess_bar(_res_sum, _res_sum_UL)
            #print("\r", end="")

        if chisq==True:
            res=_res_sum
        else:
            res=_res


        return res


    def get_chisq(self):
        model = self.model.fit_model.eval(nu=self.model.data['x'],
                                fill_SED=False,
                                get_model=True,
                                loglog=self.model.loglog)

        _res_sum, _res, _res_sum_UL = _eval_res(self.model.data['y'],
                                            model,
                                            self.model.data['dy'],
                                            self.model.data['UL'],
                                            use_UL=self.use_UL)
        return _res_sum

def _eval_res(data, model, data_error, UL, use_UL=False):

    res_no_UL = (data[~UL] - model[~UL]) / (data_error[~UL])
    res = (data - model) / (data_error)
    res_UL_log = [0]
    #print('UL.sum() ',UL.sum(),use_UL)
    if UL.sum() > 0 and use_UL is True:

        res_UL_log, res_UL = _eval_res_UL(data[UL], model[UL], data_error[UL])
        #print('ress_UL', res_UL)
        res[UL]=res_UL

    res_sum_UL= -2.0*np.sum(res_UL_log)
    res_sum=np.sum(res_no_UL * res_no_UL) + res_sum_UL
    return res_sum,res,res_sum_UL

def _eval_res_UL(y_UL,y_model,y_err):
    y = (y_UL - y_model) / (np.sqrt(2) *y_err)
    res_ul_chi2 = 0.5 * (1.0 + sp.special.erf(y))
    res_ul_chi2[res_ul_chi2 < 1E-300] = 1E-300

    res_ul = np.copy(y)
    res_ul[y > 0] = 0

    return  np.log(res_ul_chi2), res_ul



class LSBMinimizer(Minimizer):

    def __init__(self, model):

        super(LSBMinimizer, self).__init__(model)
        self.xtol=5.0E-8
        self.ftol = 5.0E-8
        self.factor=100.

    def _fit(self, max_ev,):
        #if self.use_UL is True:
        #    raise  RuntimeError('lsb minimizer currently is not supporting UL')

        bounds = [(par.fit_range_min, par.fit_range_max) for par in self.model.fit_par_free]
        max_nfev = 0 if (max_ev == 0 or max_ev == None) else max_ev
        pout, covar, info, mesg, success = leastsqbound(self.residuals_Fit,
                                                        self.model.pinit,
                                                        args=(self.model.fit_par_free,
                                                              self.model.data,
                                                              self.model.fit_model,
                                                              self.model.loglog,
                                                              False,
                                                              self.use_UL,
                                                              self.silent,),
                                                        xtol=self.xtol,
                                                        ftol=self.ftol,
                                                        factor=self.factor,
                                                        full_output=1,
                                                        bounds=bounds,
                                                        maxfev=max_nfev)

        self.mesg = mesg
        self.covar = covar
        #self.chisq =  sum(info["fvec"] * info["fvec"])

        self.pout = pout

    #def get_chisq(self):
    #    return self.chisq


class LSMinimizer(Minimizer):

    def __init__(self,model ):
        super(LSMinimizer, self).__init__(model)


    def _fit(self,max_ev):
        bounds = ([-np.inf if par.fit_range_min is None else par.fit_range_min for par in self.model.fit_par_free],
                  [np.inf if par.fit_range_max is None else par.fit_range_max for par in self.model.fit_par_free])

        max_nfev = None if (max_ev == 0 or max_ev is None) else max_ev
        fit = least_squares(self.residuals_Fit,
                            self.model.pinit,
                            args=(self.model.fit_par_free,
                                  self.model.data,
                                  self.model.fit_model,
                                  self.model.loglog,
                                  False,
                                  False,
                                  self.silent),

                            xtol=1.0E-8,
                            ftol=1.0E-8,
                            jac='3-point',
                            loss='linear',
                            f_scale=0.01,
                            bounds=bounds,
                            max_nfev=max_nfev,
                            verbose=0)
        #print(vars(fit))
        self.covar=np.linalg.inv(2 * np.dot(fit.jac.T, fit.jac))
        self.chisq=sum(fit.fun * fit.fun)
        self.mesg = fit.message
        self.pout = fit.x

    #def get_chisq(self):
    #    return self.chisq




class MinuitMinimizer(Minimizer):

    def __init__(self,model):
        if minuit_installed==True:
            pass
        else:
            raise RuntimeError('iminuit non installed')

        super(MinuitMinimizer, self).__init__(model)



    def _fit(self,max_ev=None):
        bounds = [(par.fit_range_min, par.fit_range_max) for par in self.model.fit_par_free]
        self._set_minuit_func(self.model.pinit, bounds)
        #print('=>,')
        if max_ev is None or max_ev==0:
            max_ev =10000

        self.mesg=self.minuit_fun.migrad(ncall=max_ev)
        #print('==> mesg',self.mesg)
        if iminuit.__version__ < "2":
            self.pout=[self.minuit_fun.values[k] for k in self.minuit_fun.values.keys()]
            self.p=[self.minuit_fun.values[k] for k in self.minuit_fun.values.keys()]
        else:
            self.pout=[None]*len(self.minuit_fun.values)
            self.p = [None] * len(self.minuit_fun.values)
            for ID,val in enumerate(self.minuit_fun.values):
                self.pout[ID] = val
                self.p[ID] = val

    def _set_fit_errors(self):
        if iminuit.__version__ < "2":
            self.errors = [self.minuit_fun.errors[k] for k in self.minuit_fun.errors]
        else:
            self.errors = [None] * len(self.minuit_fun.values)
            for ID,err in enumerate(self.minuit_fun.errors):
                self.errors[ID] = err





    def _set_minuit_func(self, p_init, bounds,p_error=None):
        #print('=>Hi')
        if p_error==None:
            p_error=[0.1]*len(p_init)

        p_names = ['par_{}'.format(_) for _ in range(len(p_init))]
        error_names = ['error_par_{}'.format(_) for _ in range(len(p_init))]
        p_bound_names = ['limit_par_{}'.format(_) for _ in range(len(p_init))]

        #This dict contains all the kw for
        #par
        kwdarg = {}
        for n, p, bn, b,en,e in zip(p_names, p_init, p_bound_names, bounds,error_names,p_error):
            kwdarg[n] = p
            kwdarg[bn] = b
            kwdarg[en] = e
            #print( kwdarg[n] ,p, kwdarg[bn] ,b, kwdarg[en] ,e)

        self.minuit_par_name_dict={}
        self.minuit_bounds_dict={}
        self.par_dict = {}
        #print('==> iminuit.__version__',iminuit.__version__)
        for ID,par in enumerate(self.model.fit_par_free):
            self.minuit_par_name_dict[par.name]=p_names[ID]
            self.minuit_bounds_dict[par.name]=bounds[ID]
            self.par_dict[par.name]=par
        if iminuit.__version__ < "2":
            self.minuit_fun = iminuit.Minuit(fcn=self.chisq_func,
                name=p_names,
                pedantic=False,
                errordef=1,
                **kwdarg)
        else:
            values=tuple(par.val for ID,par in enumerate(self.model.fit_par_free))
            #print('=> values',values)
            self.minuit_fun = iminuit.Minuit(self.chisq_func,
                                             values,
                                             name=p_names)
            self.minuit_fun.limits=[ bounds[ID] for ID,par in enumerate(self.model.fit_par_free) ]
            #print('=>  self.minuit_fun.limits',  self.minuit_fun.limits)
            self.minuit_fun.errordef = Minuit.LEAST_SQUARES
    def chisq_func(self, *p):
        if iminuit.__version__ < "2":
            self.p = p
        else:
            self.p = p[0]

        #print('==>p',self.p)
        #print('==>fit_par_fre', self.model.fit_par_free)
        return self.residuals_Fit(self.p,
                          self.model.fit_par_free,
                          self.model.data,
                          self.model.fit_model,
                          self.model.loglog,
                          chisq=True,
                          use_UL=self.use_UL,
                          silent=self.silent)



    #def get_chisq(self):
    #    #print ('p',self.p)
    #    return self.chisq_func(*self.p)

    def minos_errors(self,par=None):
        try:
            if par is not None:
                par=self.minuit_par_name_dict[par]
            self.calls = 0
            self.minuit_fun.minos(var=par)
            self.minuit_fun.print_param()
            self.asymm_errors = [(self.minuit_fun.merrors[(k,1.0)],self.minuit_fun.merrors[(k,-1.0)]) for k in self.minuit_fun.errors.keys()]
            self.model.reset_to_best_fit()
        except:
            raise RuntimeWarning('Fit quality not sufficient to run Minos')

    def profile(self,par,bound=2,subtract_min=True):
        self._force_silent()
        try:
            self.calls = 0
            bound=self._set_bounds(par,bound=bound)

            x, y =  self.minuit_fun.profile(self.minuit_par_name_dict[par],
                                            bound=bound,
                                            subtract_min=subtract_min)

            self.model.reset_to_best_fit()
            self._restore_silent_state()
            return x,y
        except:
            self._restore_silent_state()
            raise RuntimeWarning('Fit quality not sufficient to run profile')

    def mnprofile(self,par,bound=2,subtract_min=True):
        """

        Parameters
        ----------
        par
        bound
        subtract_min

        Returns
        -------

        """
        try:
            self._force_silent()
            self.calls = 0
            bound = self._set_bounds(par, bound=bound)
            x, y,r =  self.minuit_fun.mnprofile(self.minuit_par_name_dict[par],
                                              bound=bound,
                                              subtract_min=subtract_min)
            #print('-->r',r)
            self.model.reset_to_best_fit()
            self._restore_silent_state()
            return x,y,r
        except:
            self._restore_silent_state()
            raise RuntimeWarning('Fit quality not sufficient to run mnprofile')


    def draw_mnprofile(self,par,bound=2):
        """

        Parameters
        ----------
        par
        bound

        Returns
        -------

        """

        try:
            self._force_silent()
            self.calls = 0
            bound = self._set_bounds(par, bound=bound)
            x,y,r=self.mnprofile(par,bound,subtract_min=True)
            fig,ax = self._draw_profile(x, y, par)
            self.model.reset_to_best_fit()
            self._restore_silent_state()
            return x, y, fig,ax

        except:
            self._restore_silent_state()
            raise RuntimeWarning('Fit quality not sufficient to run draw_mnprofile')

    def draw_profile(self,par,bound=2):
        try:
            self._force_silent()
            self.calls = 0
            bound = self._set_bounds(par, bound=bound)
            x, y = self.profile(par, subtract_min=True, bound=bound)
            fig,ax=self._draw_profile(x,y,par)
            self.model.reset_to_best_fit()
            self._restore_silent_state()
            return x,y,fig,ax
        except:
            self._restore_silent_state()
            raise RuntimeWarning('Fit quality not sufficient to run draw_profile')

    def _draw_profile(self,x,y,par):
        self.calls = 0
        x = np.array(x)
        y = np.array(y)

        fig, ax = plt.subplots()

        ax.plot(x, y)
        ax.grid(True)
        ax.set_xlabel(par)
        ax.set_ylabel('FCN')


        minpos = np.argmin(y)
        ymin = y[minpos]
        tmpy = y - ymin
        # now scan for minpos to the right until greater than one
        up = self.minuit_fun.errordef
        righty = np.power(tmpy[minpos:] - up, 2)
        right_min = np.argmin(righty)
        rightpos = right_min + minpos
        lefty = np.power((tmpy[:minpos] - up), 2)
        left_min = np.argmin(lefty)
        leftpos = left_min
        le = x[minpos] - x[leftpos]
        re = x[rightpos] - x[minpos]

        ax.axvspan(x[leftpos], x[rightpos], facecolor='g', alpha=0.5)

        plt.figtext(0.5, 0.5,
                        '%s = %7.3e ( -%7.3e , +%7.3e)' % (par, x[minpos], le, re),
                        ha='center')



        return fig,ax





    def contour(self,par_1,par_2,bound=2,bins=20,subtract_min=True):
        try:
            self.calls = 0
            if np.shape(bound)==():

                bound_1 = self._set_bounds(par_1,bound)
                bound_2 = self._set_bounds(par_2,bound)

            elif np.shape(bound)==(2,) or np.shape(bound)==(2,2):

                bound_1 = self._set_bounds(par_1, bound[0])
                bound_2 = self._set_bounds(par_2, bound[1])
            else:
                raise  RuntimeError('bound must be scalar or (2,) or (2,2)')

            bound = [bound_1, bound_2]
            self._force_silent()
            x, y, z= self.minuit_fun.contour(self.minuit_par_name_dict[par_1],
                                             self.minuit_par_name_dict[par_2],
                                             subtract_min=subtract_min,
                                             bound=bound,
                                             bins=bins)
            self.model.reset_to_best_fit()
            self._restore_silent_state()
            return x,y,z
        except:
            self._restore_silent_state()
            raise RuntimeWarning('Fit quality not sufficient to run contour')


    def draw_contour(self,par_1,par_2,bound=2,levels=np.arange(5)):
        try:
            self.calls = 0
            l=self.minuit_fun.errordef*levels
            x,y,z=self.contour(par_1,par_2,bound=bound)
            fig, ax = plt.subplots()
            CS = ax.contour(x, y, z, levels,)
            ax.clabel(CS, fontsize=9, inline=1)
            ax.set_xlabel(par_1)
            ax.set_ylabel(par_2)
            self.model.reset_to_best_fit()
            return x,y,z,fig,ax
        except:
            raise RuntimeWarning('Fit quality not sufficient to run draw_contour')

    def mncontour(self,par_1,par_2,bound=2,bins=20,subtract_min=True):
        try:
            self.calls = 0
            if np.shape(bound)==():

                bound_1 = self._set_bounds(par_1,bound)
                bound_2 = self._set_bounds(par_2,bound)

            elif np.shape(bound)==(2,) or np.shape(bound)==(2,2):

                bound_1 = self._set_bounds(par_1, bound[0])
                bound_2 = self._set_bounds(par_2, bound[1])
            else:
                raise  RuntimeError('bound must be scalar or (2,) or (2,2)')

            bound = [bound_1, bound_2]
            self._force_silent()
            x, y, z= self.minuit_fun.mncontour(self.minuit_par_name_dict[par_1],
                                             self.minuit_par_name_dict[par_2],
                                             subtract_min=subtract_min,
                                             bound=bound,
                                             bins=bins)
            self.model.reset_to_best_fit()
            self._restore_silent_state()
            return x,y,z
        except:
            self._restore_silent_state()
            raise RuntimeWarning('Fit quality not sufficient to run mcontour')


    def draw_mncontour(self,par_1,par_2,bound=2,levels=np.arange(5)):
        try:
            self.calls = 0
            l=self.minuit_fun.errordef*levels
            x,y,z=self.mncontour(par_1,par_2,bound=bound)
            fig, ax = plt.subplots()
            CS = ax.contour(x, y, z, levels,)
            ax.clabel(CS, fontsize=9, inline=1)
            ax.set_xlabel(par_1)
            ax.set_ylabel(par_2)
            self.model.reset_to_best_fit()
            return x,y,z,fig,ax
        except:
            raise RuntimeWarning('Fit quality not sufficient to run draw_mcontour')

    def _set_bounds(self,par,bound):
        p = self.par_dict[par]
        if np.shape(bound)==():
            bound = [p.best_fit_val - p.best_fit_err * bound, p.best_fit_val + p.best_fit_err * bound]

        elif np.shape(bound)==(2,):
            pass
        else:
            raise  RuntimeError('bound shape',np.shape(bound),'for par',p.name,'is wrong, has to be scalare or (2,)')
        #print('bound for par', par, ' set to', bound)
        bound = self._check_bounds(par, bound)
        #print('bound for par',par,' updated to',bound)
        return bound

    def _check_bounds(self,par,bound):
        if  self.minuit_bounds_dict[par][0] is not None:
            if bound[0]<self.minuit_bounds_dict[par][0]:
                bound[0]=self.minuit_bounds_dict[par][0]

        if self.minuit_bounds_dict[par][1] is not None:
            if bound[1]>self.minuit_bounds_dict[par][1]:
                bound[1]=self.minuit_bounds_dict[par][1]

        return bound

def fit_SED(fit_model, sed_data, nu_fit_start, nu_fit_stop, fitname=None, fit_workplace=None, loglog=False, silent=False,
            get_conf_int=False, max_ev=0, use_fake_err=False, minimizer='lsb', use_UL=False,repeat=3):

    mm = ModelMinimizer(minimizer)
    return mm,mm.fit(fit_model,
                  sed_data,
                  nu_fit_start,
                  nu_fit_stop,
                  fitname=fitname,
                  fit_workplace=fit_workplace,
                  loglog=loglog,
                  silent=silent,
                  get_conf_int=get_conf_int,
                  max_ev=max_ev,
                  use_fake_err=use_fake_err,
                  use_UL=use_UL,
                  repeat=repeat)



def fit_XY(fit_model, data, x_fit_start,x_fit_stop,fitname=None,silent=False,get_conf_int=False,minimizer='minuit',
           use_UL=False, repeat=1):

    mm = ModelMinimizer(minimizer)
    return mm, mm.fit(fit_model,
                      data,
                      nu_fit_start=x_fit_start,
                      nu_fit_stop=x_fit_stop,
                      fitname=fitname,
                      fit_workplace=None,
                      loglog=False,
                      silent=silent,
                      get_conf_int=get_conf_int,
                      max_ev=0,
                      use_fake_err=False,
                      use_UL=use_UL,
                      repeat=repeat)


__author__ = "Andrea Tramacere"

import os
try:
    from gammapy.modeling.models import (
    SpectralModel,
)
    from gammapy.modeling.parameter import Parameter,Parameters
    from gammapy.estimators import FluxPoints
    from gammapy.datasets import FluxPointsDataset
    from gammapy.modeling import Fit

    gammapy_installed = True

except:
    on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
    
    if on_rtd is True:
        pass
    else:
        raise  ImportError('to use gammapy plugin you need to install gammapy: https://docs.gammapy.org/0.19/getting-started/install.html')

import astropy.units as u
import  numpy as np

def GammapyJetsetModelFactory(jetmodel,clone=True):
    
    if clone is True:
        _jetset_model = jetmodel.clone()
    else:
        _jetset_model= jetmodel
    

    class GammapyJetsetModel(SpectralModel):
        """    """

        for ID,p in enumerate(_jetset_model.parameters.par_array):
            #print(p.name)
            exec("%s=Parameter(p.name,p.val, frozen=p.frozen)"%p.name)
            if _jetset_model.parameters.par_array[ID].units is not None:
                try:
                    exec("%s.unit = '%s'"%(p.name,_jetset_model.parameters.par_array[ID].units))
                except:
                    exec("%s.unit = '' "%(p.name))
            else:
                exec("%s.unit = '' "%(p.name))

            if _jetset_model.parameters.par_array[ID].val_min is not None:
                exec("%s.min = %s"%(p.name,_jetset_model.parameters.par_array[ID].val_min))

            if _jetset_model.parameters.par_array[ID].val_max is not None:
                exec("%s.max = %s"%(p.name,_jetset_model.parameters.par_array[ID].val_max)) 
       
        def evaluate(self,energy=None,**kwargs):

            if energy is None:
                el1=np.log10( self._jetset_model.nu_min)
                el2=np.log10( self._jetset_model.nu_max)
                energy=(np.logspace(el1,el2,self._jetset_model.nu_size)*u.Hz).to('eV',equivalencies=u.spectral())
            
            nu = energy.to("Hz", equivalencies=u.spectral())

            for p in self.parameters:
                if p.name not in kwargs.keys():
                    self._jetset_model.set_par(p.name ,val=p.value)

            for k,v in kwargs.items():
                self._jetset_model.set_par(k,val=v.value)

            self._jetset_model.eval(nu=nu.value)
            _spec= self._jetset_model.spectral_components.Sum.SED.nuFnu.to('eV cm-2 s-1')/(energy.to('eV')**2)
            return _spec.to("1 / (cm2 eV s)")
        
        @property
        def jetset_model(self):
            return self._jetset_model
    
    gammapy_obj=GammapyJetsetModel()
    setattr(gammapy_obj,'_jetset_model',_jetset_model)
    setattr(gammapy_obj,'tag',_jetset_model.name)
    return gammapy_obj


import  pytest
from .base_class import TestBase

class TestEmitters(TestBase):

    def integration_suite(self,plot=False):
        self._all(plot=plot)

    def test_custom_emitters(self,plot=True):
        from jetset.jet_model import Jet

        from jetset.jet_emitters import EmittersDistribution
        import numpy as np
        def distr_func_bkn(gamma_break, gamma, s1, s2):
            return np.power(gamma, -s1) * (1. + (gamma / gamma_break)) ** (-(s2 - s1))

        n_e = EmittersDistribution('custom_bkn',spectral_type='bkn')
        n_e.add_par('gamma_break', par_type='turn-over-energy', val=1E3, vmin=1., vmax=None, unit='lorentz-factor')
        n_e.add_par('s1', par_type='LE_spectral_slope', val=2.5, vmin=-10., vmax=10, unit='')
        n_e.add_par('s2', par_type='LE_spectral_slope', val=3.2, vmin=-10., vmax=10, unit='')
        n_e.set_distr_func(distr_func_bkn)
        n_e.parameters.show_pars()
        n_e.parameters.s1.val = 2.0
        n_e.parameters.s2.val = 3.5
        if plot is True:
            n_e.plot()

        my_jet= Jet(emitters_distribution=n_e)
        my_jet.Norm_distr = True
        my_jet.parameters.N.val = 5E4
        my_jet.eval()
        np.testing.assert_allclose(my_jet.emitters_distribution.eval_N(), my_jet.parameters.N.val, rtol=1E-5)
        print(my_jet.emitters_distribution.eval_N(), my_jet.parameters.N.val)
        print(n_e.eval_N(), my_jet.parameters.N.val)
        assert (my_jet.emitters_distribution.emitters_type=='electrons')
        my_jet.save_model('test_jet_custom_emitters.pkl')
        my_jet = Jet.load_model('test_jet_custom_emitters.pkl')
        my_jet.eval()


    def test_custom_emitters_log_values(self,plot=True):
        from jetset.jet_model import Jet

        from jetset.jet_emitters import EmittersDistribution
        import numpy as np
        def distr_func_bkn(gamma_break, gamma, s1, s2):
            return np.power(gamma, -s1) * (1. + (gamma / gamma_break)) ** (-(s2 - s1))

        n_e = EmittersDistribution('custom_bkn',spectral_type='bkn',log_values=True)
        n_e.add_par('gamma_break', par_type='turn-over-energy', val=3, vmin=np.log10(2), vmax=np.log10(1E6), unit='lorentz-factor',log=True)
        n_e.add_par('s1', par_type='LE_spectral_slope', val=2.5, vmin=-10., vmax=10, unit='')
        n_e.add_par('s2', par_type='LE_spectral_slope', val=3.2, vmin=-10., vmax=10, unit='')
        n_e.set_distr_func(distr_func_bkn)
        n_e.parameters.show_pars()
        n_e.parameters.s1.val = 2.0
        n_e.parameters.s2.val = 3.5
        if plot is True:
            n_e.plot()

        my_jet= Jet(emitters_distribution=n_e)
        my_jet.Norm_distr = True
        my_jet.parameters.N.val = 5E4
        my_jet.eval()
        np.testing.assert_allclose(my_jet.emitters_distribution.eval_N(), my_jet.parameters.N.val, rtol=1E-5)
        print(my_jet.emitters_distribution.eval_N(), my_jet.parameters.N.val)
        print(n_e.eval_N(), my_jet.parameters.N.val)
        assert (my_jet.emitters_distribution.emitters_type=='electrons')
        my_jet.save_model('test_jet_custom_emitters.pkl')
        my_jet = Jet.load_model('test_jet_custom_emitters.pkl')
        my_jet.eval()

    def test_emitters_log_values(self,plot=True):
        from jetset.jet_model import Jet

        from jetset.jet_emitters_factory import  EmittersFactory
        ef=EmittersFactory()
        for distr_name in EmittersFactory.available_distributions_list():
            n_e=ef.create_emitters(name=distr_name,log_values=True)
            my_jet = Jet(emitters_distribution=n_e)
            my_jet.eval()
            my_jet.save_model('test_jet_custom_emitters_log.pkl')
            my_jet = Jet.load_model('test_jet_custom_emitters_log.pkl')
            my_jet.eval()

    def test_custom_emitters_array(self,plot=True):
        from jetset.jet_model import Jet
        from jetset.jet_emitters import EmittersArrayDistribution
        import numpy as np

        # gamma array
        gamma = np.logspace(1, 8, 500)

        # gamma array this is n(\gamma) in 1/cm^3/gamma
        n_gamma = gamma ** -2 * 1E-5 * np.exp(-gamma / 1E5)

        N1 = np.trapz(n_gamma, gamma)

        n_distr = EmittersArrayDistribution(name='array_distr', emitters_type='protons', gamma_array=gamma, n_gamma_array=n_gamma,normalize=False)

        N2 = np.trapz(n_distr._array_n_gamma, n_distr._array_gamma)

        j = Jet(emitters_distribution=n_distr, verbose=False)

        j.parameters.z_cosm.val = z = 0.001
        j.parameters.beam_obj.val = 1
        j.parameters.N.val = 1
        j.parameters.NH_pp.val = 1
        j.parameters.B.val = 0.01
        j.parameters.R.val = 1E18
        j.set_IC_nu_size(100)
        j.gamma_grid_size = 200

        N3 = np.trapz(j.emitters_distribution.n_gamma_p, j.emitters_distribution.gamma_p)

        np.testing.assert_allclose(N1, N2, rtol=1E-5)
        np.testing.assert_allclose(N1, N3, rtol=1E-2)
        np.testing.assert_allclose(N1, j.emitters_distribution.eval_N(), rtol=1E-2)
        assert (j.emitters_distribution.emitters_type == 'protons')

        j.eval()
        j.save_model('test_jet_custom_emitters_array.pkl')
        j = Jet.load_model('test_jet_custom_emitters_array.pkl')
        j.eval()


    def test_emitters_U(self,plot=True):
        from jetset.jet_model import Jet
        from jetset.jetkernel import jetkernel
        import numpy as np

        j = Jet(emitters_distribution='plc', verbose=False, emitters_type='protons')
        j.parameters.z_cosm.val = z = 0.001
        j.parameters.beam_obj.val = 1
        j.parameters.gamma_cut.val = 1000 / (jetkernel.MPC2_TeV)
        j.parameters.NH_pp.val = 1
        j.parameters.N.val = 1
        j.parameters.p.val = 2.0
        j.parameters.B.val = 1.0
        j.parameters.R.val = 1E18
        j.parameters.gmin.val = 1
        j.parameters.gmax.val = 1E8
        j.set_emiss_lim(1E-60)
        j.set_IC_nu_size(100)
        j.gamma_grid_size = 200
        gmin=1.0/jetkernel.MPC2_TeV
        j.set_N_from_U_emitters(1.0, gmin=gmin)
        m = j.emitters_distribution.gamma_p>gmin
        N1 = jetkernel.MPC2*np.trapz(j.emitters_distribution.n_gamma_p[m]*j.emitters_distribution.gamma_p[m],j.emitters_distribution.gamma_p[m])
        N2 = j.emitters_distribution.eval_U(gmin=gmin)
        np.testing.assert_allclose(N1, N2, rtol=1E-5)



__author__ = "Andrea Tramacere"

import numpy as np
from .utils import *

__all__=['convert_nu_to_blob','convert_nu_to_src','convert_nuFnu_to_nuLnu_src']

def convert_nu_to_blob(nu,in_frame,delta,z):
    """converts Energies/frequencies from in_frame  to blob rest frame


        Args:
            delta: beaming factor
            z: redshift
            in_frame: startign reference frame, obs=observer (earth), src=corrected for redshift, blob=corrected for redshift and beaming

        Returns:
            E/nu in blob rest frame

    """
    check_frame(frame=in_frame)

    if in_frame=="obs":
        z_c=z
        delta_c=delta
    elif in_frame=="src":
        z_c=0
        delta_c=delta
    elif in_frame=="blob":
        delta_c=1
        z_c=0
    else:
        unexpected_behaviour()

    return nu*(1+z_c)/delta_c




def convert_nu_to_src(nu,z,in_frame):
    """converts Energies/frequencies from in_frame  to blob rest frame


        Args:
            delta: beaming factor
            z: redshift
            in_frame: startign reference frame, obs=observer (earth), src=corrected for redshift, blob=corrected for redshift and beaming
    

        Returns:
            E/nu in blob rest frame

    """
    check_frame(in_frame)
    if in_frame=="obs":
        z_c=z
    elif in_frame=="src":
        z_c=0
    else:
         unexpected_behaviour()

    return nu*(1+z_c)


def convert_nuFnu_to_nuLnu_src(nuFnu,z,in_frame,dl):
    check_frame(in_frame)
    if in_frame == "obs":
       c=4.0*np.pi*dl**2
    elif in_frame == "src":
        c=1.0
    else:
        unexpected_behaviour()

    return c*nuFnu


def convert_nuLnu_src_to_nuLnu_blob(nuLnu,beaming,in_frame):
    check_frame(in_frame)
    c=1/(beaming**4)
    return c*nuLnu


def convert_nu_src_to_nu_blob(nu_src,beaming,in_frame):
    check_frame(in_frame)
    return nu_src/beaming

    
    
    


"""
JetSeT package
"""


import pkgutil
import os
#import json
from .utils import get_info

from . import jetkernel

if jetkernel._refactored_ is True:
    pass
else:
    raise RuntimeError('please, uninstall the current version of jetekernel')

__author__ = "Andrea Tramacere"


pkg_dir = os.path.abspath(os.path.dirname(__file__))
pkg_name = os.path.basename(pkg_dir)
__all__ = []

_info = get_info()
__version__ = _info['version']

if 'label'  in _info.keys():
    __label__= _info['label']
else:
    __label__= None

for importer, modname, ispkg in pkgutil.walk_packages(path=[pkg_dir],
                                                      prefix=pkg_name+'.',
                                                      onerror=lambda x: None):
    if ispkg is True:
        __all__.append(modname)
    else:
        pass


data_dir = os.path.dirname(__file__)+'/data'




__author__ = "Andrea Tramacere"


import numpy as np
import copy


from astropy.table  import  Table,Column
from astropy import  units as u
from astropy.units import cds
import  pickle
from .plot_sedfit import PlotSED, plt

from .output import section_separator
from .utils import *
from .cosmo_tools import Cosmo
from .frame_converter import convert_nu_to_src, convert_nuFnu_to_nuLnu_src

cds.enable()

__all__=['get_data_set_msk','get_freq_range_msk','lin_to_log','log_to_lin','ObsData','Data']


class Data(object):
    """
    Class to store obs data
    """
    def __init__(self,
                 data_table=None,
                 n_rows=None,
                 meta_data=None,
                 import_dictionary=None,
                 cosmo=None):
        """_summary_

        Parameters
        ----------
        data_table : astropy table, optional
            _description_, by default None
        n_rows : int, optional
            the number of rows for an empty table if no data_table is passed, by default None
        meta_data : dictionary, optional
            the dictionary for the metadata, by default None
        import_dictionary : dictionary, optional
            a dictionary mapping the column renaming if needed for the data_table, by default None
        cosmo : jetset.cosmo_tools.Cosmo, optional
            the cosmology object, by default None
        """

        if cosmo is None:

            self.cosmo = Cosmo()
        else:

            self.cosmo = cosmo

        self._necessary_meta = ['data_scale', 'z', 'restframe']
        self._allowed_meta={}
        self._allowed_meta['z']= None
        self._allowed_meta['UL_CL'] = None
        self._allowed_meta['restframe'] = ['obs','src']
        self._allowed_meta['data_scale'] = ['lin-lin','log-log']
        self._allowed_meta['obj_name'] = None

        self._names = ['x', 'dx', 'y', 'dy', 'T_start', 'T_stop', 'UL', 'data_set']
        self._dt = ('f8', 'f8', 'f8', 'f8', 'f8', 'f8', 'bool', 'S16')

        self._units = [u.Hz, u.Hz,(u.erg / (u.cm ** 2 * u.s)) ,(u.erg / (u.cm ** 2 * u.s)), cds.MJD, cds.MJD, None, None]

       
        if isinstance(data_table,str):
            data_table = Table.read(data_table, guess=True)

        if data_table is None:
            self._build_empty_table(n_rows,meta_data=meta_data)

        else:
            if import_dictionary is not None:
                for k in import_dictionary.keys():
                    data_table.rename_column(k, import_dictionary[k])

            self._table = copy.deepcopy(data_table)

        if meta_data is not None:
            self._table.meta = {}
            for k in meta_data.keys():

                self.set_meta_data(k,meta_data[k])
        else:
            for k in self._table.meta.keys():
                self.set_meta_data(k,self._table.meta[k])

        self._check_table()
        self._check_frame_and_scale()
        self._convert_units()

        #self._table.sort('x')


    @property
    def table(self):
        return self._table

    @property
    def metadata(self):
        return self._table.meta

    @classmethod
    def from_file(cls,data_table,format='ascii.ecsv',import_dictionary=None,guess=None):
        return cls(data_table= Table.read(data_table, format=format,guess=guess),import_dictionary=import_dictionary)

    def save_file(self, name, format='ascii.ecsv'):
        self._table.write(name,format=format,overwrite=True)



    def _check_frame_and_scale(self):
        if self.metadata['data_scale'] == 'log-log':
            self._table['x'], self._table['dx'] = log_to_lin(self._table['x'], self._table['dx'])
            self._table['y'], self._table['dy'] = log_to_lin(self._table['y'], self._table['dy'])

            self.metadata['data_scale']='lin-lin'

        if self.metadata['restframe'] == 'src':
            if self._table['y'].unit.is_equivalent('erg/s') and  self._table['dy'].unit.is_equivalent('erg/s'):
                pass
            else:
                raise  RuntimeError('when importing src frame table, units for luminosities have to be in erg/s')



            _c=self.cosmo.get_DL_cm(self.metadata['z'])
            _c=1.0/(4*np.pi*_c*_c)
            self._table['y']  = self._table['y'] * _c
            self._table['dy'] = self._table['dy']  * _c
            self._table['y'].unit = 'erg/(cm2 s)'
            self._table['dy'].unit = 'erg/(cm2 s)'
            self.metadata['restframe'] = 'obs'

    def _check_table(self):

        for ID,_cn in enumerate(self._names):
            if _cn not in self._table.colnames:
                self._table.add_column(index=ID,col=Column(name=_cn,dtype=self._dt[ID],unit=self._units[ID],data=np.zeros(len(self._table))))

        for _n in  self._necessary_meta:
            if _n not in  self._table.meta:
                raise RuntimeError('meta data',_n,'not defined, but it is necessary')


    def _convert_units(self):
        for ID,_cn in enumerate(self._names):
            if hasattr(self._table[_cn],'unit'):
                if self._table[_cn].unit is not None and self._units[ID] is not None:
                    try:
                        self._table[_cn]=self._table[_cn].to(self._units[ID])
                    except Exception as e:
                        try:
                            self._table[_cn] = self._table[_cn].to(self._units[ID], equivalencies=u.spectral())
                        except Exception as e:
                            raise RuntimeError('Unit conversion problem for ',self._table[_cn].unit,'to',self._units[ID],repr(e))


    def set_meta_data(self,m,v):
        if m not in self._allowed_meta:
            raise RuntimeError('meta data ',m,'not in allowed',self._allowed_meta.keys())

        if  self._allowed_meta[m] is not None:
            if v not in self._allowed_meta[m]:
                raise RuntimeError('meta data value ', v, 'not in allowed', self._allowed_meta[m])


        self._table.meta[m]=v



    def set_field(self,field,value,unit=None):

        if unit is None:
            if hasattr(self._table[field], 'unit'):
                unit = self._table[field].unit

        self._table[field] = value
        if unit is not None:
            self._table[field].unit = unit

        

    def _build_empty_table(self,n_rows,meta_data=None):


        self._table = Table(np.zeros((n_rows, len(self._names))), names=self._names, dtype=self._dt)
        for ID,c in enumerate(self._table.columns):
            if self._units[ID] is not None:
                #print(ID,c, units[ID])
                self._table[c]*=self._units[ID]

        self._table.meta['z'] = 0
        self._table.meta['UL_CL'] = 0.95
        self._table.meta['restframe'] = 'obs'
        self._table.meta['data_scale'] = 'lin-lin'
        self._table.meta['obj_name'] = 'new-src'

        if meta_data is not None:
            for k in self._table.meta.keys():
                if k in meta_data.keys():
                    self._table.meta[k]=meta_data[k]



    @classmethod
    def from_asdc(cls, asdc_sed_file, obj_name, z, restframe, data_scale):
        with open(asdc_sed_file, 'r') as f:
            lines = f.readlines()
        # print(len(lines),type(lines),lines)
        for l in lines[:]:
            if l.startswith('#'):
                lines.remove(l)

        UL = np.zeros(len(lines), dtype=np.bool)

        for ID, l in enumerate(lines):
            t = l.strip().split(';')

            if len(t) > 1:
                # print(t[1])

                if 'UPPER LIMIT' in t[1]:
                    UL[ID] = True
                    lines[ID] = t[0]

        d = np.genfromtxt(lines)
        d = np.column_stack((d, UL))

        data_table = Table(d, names=['x', 'dx', 'y', 'dy', 'T_start', 'T_stop', 'UL'])
        data_table['x'] = data_table['x'] * u.Hz
        data_table['dx'] = data_table['dx'] * u.Hz
        data_table['y'] = data_table['y'] * (u.erg / (u.cm ** 2 * u.s))
        data_table['dy'] = data_table['dy'] * (u.erg / (u.cm ** 2 * u.s))
        data_table['T_start'] = data_table['T_start'] * cds.MJD
        data_table['T_stop'] = data_table['T_stop'] * cds.MJD
        data_table['UL']=np.array(data_table['UL'], dtype=np.bool)
        data_table.meta['z'] = z
        data_table.meta['restframe'] = restframe
        data_table.meta['data_scale'] = data_scale
        data_table.meta['obj_name'] = obj_name
        return cls(data_table=data_table)


class ObsData(object):
    """ObsData class

    This class provides a  powerful interface  to load observational data stored in a file.
    The following parameters set the corresponding class members
    """


    def __init__(self,
                 cosmo=None,
                 data_table=None,
                 dupl_filter=False,
                 data_set_filter=None,
                 UL_filtering=False,
                 UL_value=None,
                 UL_CL=0.95,

                 **keywords):
        
        """

        Parameters
        ----------
        cosmo
        data_table
        dupl_filter
        data_set_filter
        UL_filtering
        UL_value
        UL_CL
        keywords
        """
        
        self.z=None
        self.data=None
        self.obj_name=None
        self.restframe=None
        self.col_types=None
        self.col_nums=None
        self.data_scale=None
        self.UL_CL=UL_CL

        if data_set_filter is not None:
            self.data_set_filter=data_set_filter.split(',')
        else:
            self.data_set_filter=['No']

        if cosmo is None:
            if hasattr(data_table,'cosmo'):
                self.cosmo=data_table.cosmo
            else:
                self.cosmo=Cosmo()
        else:
            self.cosmo=cosmo

        self.UL_value=UL_value
        self.UL_filtering=UL_filtering
        self.zero_error_replacment=0.2
        self.fake_error=0.2



        if  hasattr(data_table,'table'):
            _t=data_table.table
        else:
            _t=data_table

        self._input_data_table=None

        if isinstance(_t,Table):
            self._input_data_table = _t
        else:
            raise RuntimeError('table is not an astropy Table')

        self.allowed_keywords = {'z': None}
        self.allowed_keywords['obj_name'] = None
        self.allowed_keywords['restframe'] = ['obs', 'src']
        self.allowed_keywords['data_scale'] = ['lin-lin', 'log-log']
        self.allowed_keywords['file_name'] = None
        self.allowed_keywords['UL_CL'] = None
        self.allowed_keywords['col_types'] = None
        self.allowed_keywords['col_nums'] = None


        _skip = ['data_table', 'n_rows']

        if self._input_data_table is not None:
            self._set_kw(self._input_data_table.meta)

        self._set_kw(keywords,_skip)

        self._build_data(dupl_filter=dupl_filter)
        self.data.sort('nu_data')

    def _set_kw(self,keywords,skip=[]):
        keys = sorted(keywords.keys())
        # print keys

        for kw in keys:
            if kw not in skip:
                if kw in self.allowed_keywords.keys():

                    if self.allowed_keywords[kw] is not None:
                        # check that the kyword value is correct
                        if keywords[kw] not in self.allowed_keywords[kw]:
                            print("keyword=%s, has wrong value=%s, allowed are %s" % (
                            kw, keywords[kw], self.allowed_keywords[kw]))
                            raise ValueError

                    setattr(self, kw, keywords[kw])

                else:

                    print("wrong keyword=%s, not in%s " % (kw, self.allowed_keywords.keys()))

                    raise ValueError

        if self.z is not None:
            self.z = float(self.z)

        if self.UL_CL is not None:
            self.UL_CL = float(self.UL_CL)

    @property
    def metadata(self,skip=['col_types','col_nums']):
        _d={}
        for k in self.allowed_keywords.keys():
            if hasattr(self,k) and k not in skip:
                _d[k]=getattr(self,k)
        return _d

    def _build_empty_table(self,n_rows=None):
        sed_dt = [('nu_data', 'f8')]
        sed_dt.append(('dnu_data', 'f8'))
        sed_dt.append(('nuFnu_data', 'f8'))
        sed_dt.append(('dnuFnu_data', 'f8'))

        sed_dt.append(('nu_data_log', 'f8'))
        sed_dt.append(('dnu_data_log', 'f8'))
        sed_dt.append(('nuFnu_data_log', 'f8'))
        sed_dt.append(('dnuFnu_data_log', 'f8'))

        sed_dt.append(('dnuFnu_fake', 'f8'))
        sed_dt.append(('dnuFnu_fake_log', 'f8'))

        sed_dt.append(('UL', 'bool'))
        sed_dt.append(('zero_error', 'bool'))
        sed_dt.append(('T_start', 'f8'))
        sed_dt.append(('T_stop', 'f8'))
        sed_dt.append(('data_set', 'S16'))

        self.dt = np.dtype(sed_dt)

        if n_rows is None:
            n_rows=len(self._input_data_table)

        return Table(rows=np.zeros((n_rows, len(sed_dt))), dtype=[d[1] for d in sed_dt],
                          names=[d[0] for d in sed_dt])

    @staticmethod
    def load(file_name,):
        _obj=pickle.load(open(file_name, "rb"))
        _obj.data.sort('nu_data')
        return _obj

    def save(self, file_name):
        self.data.sort('nu_data')
        pickle.dump(self, open(file_name, 'wb'), protocol=pickle.HIGHEST_PROTOCOL)


    @property
    def table(self):
        return self.data

    def _build_data(self,dupl_filter=False):
        """
        private method to load and build the SED data
        
        :param dupl_filter: keyword to perfrom filtering of duplicated entries
        :type dupl_filter: bool 
        :ivar dt: numpy dtype  for the SED data
         
        
        
        **Data Processing**
        
        - checks if data errors are provided
        - separates historical from simultaneosu (used in the fit) data
        - filters upper limits 
        - removes duplicate entries
        - performs restframe transformation
        - performs `lin-lin`, `log-log` transformations
        

       
       
        """


        self.data=self._build_empty_table()

        self._col_dict = {'x': 'nu_data'}
        self._col_dict['y'] = 'nuFnu_data'
        self._col_dict['dx'] = 'dnu_data'
        self._col_dict['dy'] = 'dnuFnu_data'
        self._col_dict['T_start'] = 'T_start'
        self._col_dict['T_stop'] = 'T_stop'
        self._col_dict['data_set'] = 'data_set'
        self._col_dict['UL'] = 'UL'

        self._log_col_dict = {'x': 'nu_data_log'}
        self._log_col_dict['y'] = 'nuFnu_data_log'
        self._log_col_dict['dx'] = 'dnu_data_log'
        self._log_col_dict['dy'] = 'dnuFnu_data_log'
        self._log_col_dict['T_start'] = 'T_start'
        self._log_col_dict['T_stop'] = 'T_stop'
        self._log_col_dict['data_set'] = 'data_set'
        self._log_col_dict['UL'] = 'UL'
        
        if self.data_scale=='lin-lin':
            self._file_col_dict= self._col_dict
            
        
        elif self.data_scale=='log-log':
             self._file_col_dict= self._log_col_dict
        
        else:
            raise RuntimeError('data_scale not specified')

        self.col_types = []
        for n in self._input_data_table.colnames:
            if n  in self._file_col_dict.keys():
                self.col_types.append(n)

                self.data[self._file_col_dict[n]]=self._input_data_table[n]

        self.col_nums=len(self.col_types)



        #-------------------------------------------------------------------------
        # duplicate entries 
        if dupl_filter==True:
            #print ("---> filtering for dupl entries")
            self.remove_dupl_entries(self.data)
       
        #-------------------------------------------------------------------------          

        # sets UL entries
        self.set_UL(self.UL_value)

        # set error if not present:
        self.set_fake_error(self.fake_error)





        self.set_zero_error()

        if self.UL_filtering==True:
            self.filter_UL(self.UL_value)
        
        
        if self.data_set_filter!=['No']:
            self.filter_data_set( self.data_set_filter)
       

        self._set_data_frame_and_scale()

       
        
        if self.data['dnuFnu_data'] is None and self.data_scale== 'lin-lin':
        
            self.data['dnuFnu_data']= self.data['Fnu_data'] * self.fake_error
            
            print ("Warning: error were not provided ")
            print ("         assigning %f            ")
            print ("         set error with .set_error"%self.fake_error)
            #print self.data['dnuFnu_data']
        
        
        if self.data['dnuFnu_data_log'] is None and self.data_scale== 'log-log':
        
            self.data['dnuFnu_data_log']= np.ones(self.data['nu_data_log'].size) * self.fake_error
            
            print ("Warning: error were not provided ")
            print ("         assigning %f           ")
            print ("         set error with .set_error"%self.fake_error)

    








    def _set_data_frame_and_scale(self):
        
        
        #-------------------------------------------------------------------------          
        # handles data cosmological conversion
        # and axis transformation

        check_frame(self.restframe)

        if self.restframe=='obs':
            self.nu_conv_factor=1
            self.Lum_conv_factor=1
        
        elif self.restframe=='src':
            #TODO this must be the same as in jetset
            DL=self.cosmo.get_DL_cm(self.z)
            print ("--->DL=%e"%DL)
            
            #!! usa le funzioni in frame_converter
            self.nu_conv_factor=1.0/(1+self.z)
            self.Lum_conv_factor=1.0/(np.pi*4.0*DL*DL)
        else:
            unexpected_behaviour()

        if self.data_scale=='lin-lin':
            
            if 'dy' in self.col_types :
                self.data['nuFnu_data_log'], self.data['dnuFnu_data_log']=self.lin_to_log(val=self.data['nuFnu_data'], err=self.data['dnuFnu_data'])
            else:
                self.data['nuFnu_data_log']=self.lin_to_log(val=self.data['nuFnu_data'])
                
            if 'dx' in self.col_types:
                self.data['nu_data_log'], self.data['dnu_data_log']=self.lin_to_log(val=self.data['nu_data'], err=self.data['dnu_data'])
            else:
                self.data['nu_data_log']=self.lin_to_log(val=self.data['nu_data'])
            
            self.data['nu_data_log']+= np.log10(self.nu_conv_factor)
            self.data['nuFnu_data_log']+= np.log10(self.Lum_conv_factor)
            
            
            
        if self.data_scale=='log-log':

            if 'dy' in self.col_types:
                self.data['nuFnu_data'], self.data['dnuFnu_data']=self.log_to_lin(log_val=self.data['nuFnu_data_log'], log_err=self.data['dnuFnu_data_log'])
            else:
                self.data['nuFnu_data']=self.log_to_lin(log_val=self.data['nuFnu_data_log'])
                
            if 'dx' in self.col_types:
                self.data['nu_data'], self.data['dnu_data']=self.log_to_lin(log_val=self.data['nu_data_log'], log_err=self.data['dnu_data_log'])
            else:
                self.data['nu_data']=self.log_to_lin(log_val=self.data['nu_data_log'])
            
            
            #print self.data['nuFnu_data'],self.data['dnuFnu_data']
            self.data['nu_data']*=self.nu_conv_factor
            self.data['nuFnu_data']*=self.Lum_conv_factor
        
        

    def _set_md_from_data_table(self,name=None,val=None):




        md_dic = {}
        #md_dic['z'] = None
        #md_dic['file_name'] = None
        #md_dic['obj_name'] = None
        #md_dic['restframe'] = None
        #md_dic['data_scale'] = None
        #md_dic['col_types'] = None
        #md_dic['col_nums'] = None

        if name is None:

            for k in self.allowed_keywords.keys():
                if k in self._input_data_table.meta.keys():
                    #md_dic[k]=self._input_data_table.meta[k]
                    setattr(self, k, self._input_data_table.meta[k])
                    #print(k,md_dic[k])
        else:
            if name in md_dic.keys():
                setattr(self, name, val)
            else:
                raise RuntimeError('meta name',name,'not in allowed',md_dic.keys())

        if self.z is not None:
           self.z=float(self.z)

        if self.UL_CL is not None:
            self.UL_CL = float(self.UL_CL)
    
    
    
    def filter_data_set(self,filters,exclude=False,silent=False):

        filters=filters.split(',')

        if silent is False:
            if exclude==False:
                print ("---> including  only data_set/s",filters)
            else:
                print ("---> excluding  data_set/s",filters)

        
        msk=np.ones(self.data['nu_data'].size, dtype=bool)

        if exclude == True:
            msk = np.ones(self.data['nu_data'].size, dtype=bool)

        if exclude == False:
            msk = np.zeros(self.data['nu_data'].size, dtype=bool)

        for filter in filters:
            #print ('filter',filter)
            msk1= self.data['data_set'] == filter
            if exclude == True:
                msk1 = np.invert(msk1)
                msk = np.logical_and(msk, msk1)
            else:
                msk=np.logical_or(msk,msk1)
            if silent is False:
                print('filter', filter, np.sum(msk))

        
        self.data=self.data[msk]
        if silent is False:
            print("---> data sets left after filtering",self.show_data_sets())
            print ("---> data len after filtering=%d" % len(self.data['nu_data']))
        

    def set_fake_error(self,):

        if 'dy' not in self.col_types:
            self.set_error(self.fake_error)

        
    def set_UL(self,val=None):
        
        self.UL_value=val

        if self.UL_value is not None:
            print("---> setting  UL")
            if 'dy' in self.col_types and self.data_scale=='lin-lin':
                print ("---> Settin  UL for val",val)
                error_array=self.data['dnuFnu_data']


            elif 'dy' in self.col_types and self.data_scale=='log-log':

                error_array=self.data['dnuFnu_data_log']

            else:
                error_array=None

            if error_array is not None:
                self.data['UL']= error_array < val


        
        
    def set_zero_error(self,val=0.2,replace_zero=True):
        
        self.zero_error_replacment=val
        #print("---> replacing zero error with relative error ", val)
        if 'dy' in self.col_types and self.data_scale=='lin-lin':

            error_array=self.data['dnuFnu_data']
                       
        
        elif 'dy' in self.col_types and self.data_scale=='log-log':

            error_array=self.data['dnuFnu_data_log']
        
        else:
            error_array=None

        if error_array is not None:
            #works only on data that are not UL
            self.data['zero_error']= error_array <= 0.0
            #self.data['zero_error']*=~self.data['UL']

            if replace_zero == True:
                #self.set_error(self.zero_error_replacment, data_msk=self.data['zero_error'])
                if self.data_scale=='lin-lin':
                    self.data['dnuFnu_data'][self.data['zero_error']]= self.zero_error_replacment * self.data['nuFnu_data'][self.data['zero_error']]

                if self.data_scale=='log-log':
                    self.data['dnuFnu_data_log'][self.data['zero_error']]= self.zero_error_replacment / np.log(10)

        
        
    def filter_UL(self,val=None):
        """
        remove the  upper limits points from  from data
        
        :param val: minimum value to set the upper limit. **As default, negative errors indicates upper limits, hence val=0.**
        :Retruns msk: a boolean array to mask the upper limits, i.e. all the data points with negative errors.
        
        """
        
#         
#         if 'dy' in self.col_types and self.data_scale=='lin-lin':
#             print "---> filtering for UL"
#             error_array=self.data['dnuFnu_data']
#             
#             msk= error_array<val
#             if len(msk)>0:
#                 self.ULnu_data=self.data['nu_data'][msk]
#                 self.UL_data=self.data['nuFnu_data'][msk]
#         
#         if 'dy' in self.col_types and self.data_scale=='log-log':
#             print "---> filtering for UL"
#             erro_array=self.data['dnuFnu_data_log']
#             msk= error_array<val
#             if len(msk)>0:
#                 self.UL_nu_data_log=self.data['nu_data_log'][msk]
#                 self.UL_data_log=self.data['nuFnu_data_log'][msk]
#             
#         
#         
        if val is None:
            val=self.UL_value
            
        print ("---> filtering  UL for val",val)
        if 'dy' in self.col_types and self.data_scale=='lin-lin':   
            error_array=self.data['dnuFnu_data']
        
           
        
        if 'dy' in self.col_types and self.data_scale=='log-log':
            error_array=self.data['dnuFnu_data_log']
        
        
        msk= error_array<val
        msk=np.invert(msk)
        
        self.data=self.data[msk]
        print ("---> data len after filtering=%d" % len(self.data))
        
     
    def filter_time(self,T_min=None,T_max=None,exclude=False):
        """
        filter the data, keeping all the data with
        T_min <T< T_max if exclude=False (defualt).
        The opposite if exclude=True
        
        
        :param T_min: lower limit of the range (MJD)
        :type T_min: float
        :param T_max: upper limit of the range (MJD)
        :type T_max: float
        """
        
        
        msk1=np.ones(self.data['nu_data'].size, dtype=bool)
        msk2=np.ones(self.data['nu_data'].size, dtype=bool)
        if T_min is not None:
            msk1= self.data['T_start'] >= T_min
        
        if T_max is not None:
            msk2= self.data['T_stop'] <= T_max
            
        msk=msk1*msk2
        if exclude==True:
            msk=np.invert(msk)
        
        self.data=self.data[msk]
        print ("---> data len after filtering=%d" % len(self.data))
        
        
    def filter_freq(self,nu_min=None,nu_max=None,exclude=False):
        """
        filter the data, keeping all the data with
        nu_min <nu< nu_max if exclude=False (defualt).
        The opposite if exclude=True
        
        
        **both nu_max and nu_min are in linear scale**
        
        :param nu_min: lower limit of the range (linear scale, in Hz)
        :type nu_min: float
        :param nu_max: upper limit of the range (linear scale, in Hz)
        :type nu_max: float
        """
        
        msk1=np.ones(self.data['nu_data'].size, dtype=bool)
        msk2=np.ones(self.data['nu_data'].size, dtype=bool)
        if nu_min is not None:
            msk1= self.data['nu_data'] >= nu_min
        
        if nu_max is not None:
            msk2= self.data['nu_data'] <= nu_max
            
        msk=msk1*msk2
        
        if exclude==True:
            msk=np.invert(msk)
        self.data=self.data[msk]

        print ("---> data len after filtering=%d" % len(self.data))
    
    def reset_data(self):
        self._build_data()
       
  
  
    def remove_dupl_entries(self,data):
        """
         remove duplicate entries
        
        :param data: (array) 2-dim array storing the the table of the data.
        :Returns msk:  a boolean array to mask the duplicated entries
        
        .. note::
        
            One entry is flagged as duplicated as each comlum in a row of the data table
            is equal to the corresponding elements of another row
        """
        print ("---> remove duplicate entries")
        msk=np.array(np.ones(len(data)),dtype=bool)

        for i in range(1,len(data)):
            for j in range(1,len(data[:i])):
                test= data[i]==data[j]
                if test.all():
                    #print i,tests,data[i],data[j]
                    msk[i]=False
               
        data=data[msk]
            
            
            
    def group_data(self,N_bin=None,bin_width=None):
        
        """
        function to perform a spectral group of the data

        :param N_bin: (int)
        :param bin_width: (float) logarthmic
        
        .. note::
    
            To perform  a rebinning of the data has to be provided either ``N_bin`` or ``bin_width``.
        """
        
        print (section_separator)
        

        if N_bin is not None and bin_width is  not None:
            print ("you can provide either N_bin or bin_width")
            raise ValueError 
        elif N_bin is None and bin_width is None:
            print ("you must provide either N_bin or bin_width")
            raise ValueError

        #if nu_min is None:
        xmin= self.data['nu_data_log'].min() * 0.99
        #else:
        #   xmin= np.log10(nu_min)

        #if nu_max is None:
        xmax= self.data['nu_data_log'].max() * 1.01
        #else:
        #xmin = np.log10(nu_max)


        if N_bin is None:
            N_bin=int((xmax-xmin)/bin_width)

        if bin_width is None:
            bin_width=(xmax-xmin)/N_bin


        bin_grid=np.linspace(xmin,xmax,N_bin)
        
        
        print ("***  binning data  ***")
        print ("---> N bins=",N_bin)
        print ("---> bin_widht=",bin_width)
  
        self.data_reb=self._build_empty_table(n_rows=bin_grid.size)

    
        x_bin=np.zeros(bin_grid.size)
        y_bin=np.zeros(bin_grid.size)
        dx_bin=np.zeros(bin_grid.size)
        dy_bin=np.zeros(bin_grid.size)    

        #gives the id of element falling in each bin
        bin_elements_id=np.digitize(self.data['nu_data_log'], bin_grid)
        
        
        for id in range(bin_grid.size):
            msk1=[bin_elements_id==id][0]
            #Remove UL
            msk2=np.invert(self.data['UL'])
            msk=msk1*msk2
            
            if msk.any()==True>0:
                if len(self.data['dnuFnu_data_log'][msk])>1:
                    w=1.0/self.data['dnuFnu_data_log'][msk]
                    #print "w",self.data['dnuFnu_data'][msk]
                    w=w*w
                    
                    #w=1/sig_i^2
                    y_bin[id],sum_w=np.average(self.data['nuFnu_data_log'][msk], axis=0, weights=w, returned=True)
                    V1=sum_w
                    V2=w*w
                    V2=V2.sum()
                    V3= w * (self.data['nuFnu_data_log'][msk] - y_bin[id]) * (self.data['nuFnu_data_log'][msk] - y_bin[id])
                    V3=V3.sum()
                    
                    if V3==0 or V1*V1-V2==0:
                    
                        #print"weighted average not possible for bin=",id
                        #print"V3 V1*V1-V2", V1,V1*V1-V2
                        #print"using err=sqrt(1/sum(w))"
                        dy_bin[id]=np.sqrt(1.0/V1)
                    
                    else:
                        dy_bin[id]=np.sqrt((V1/(V1*V1-V2))*V3)
                    
                    dx_bin[id]=bin_width/2
                    x_bin[id]=bin_grid[id]-dx_bin[id]
                    #print"x", x_bin[id],len(w)
                else:
                    #print "xxx"
                    x_bin[id]=self.data['nu_data_log'][msk]
                    y_bin[id]=self.data['nuFnu_data_log'][msk]
                    dx_bin[id]=bin_width/2
                    dy_bin[id]=self.data['dnuFnu_data_log'][msk]
                    
        self.data_reb['nu_data_log']=x_bin
        self.data_reb['nuFnu_data_log']=y_bin
        self.data_reb['dnu_data_log']=dx_bin
        self.data_reb['dnuFnu_data_log']=dy_bin
        
        #remove empty bins
        msk=[self.data_reb['nu_data_log']!=0]
        
        self.data_reb=self.data_reb[msk]
        
        
         
        self.data_reb['nuFnu_data'],self.data_reb['dnuFnu_data']=self.log_to_lin(log_val=self.data_reb['nuFnu_data_log'], log_err=self.data_reb['dnuFnu_data_log'])
          
        self.data_reb['nu_data'],self.data_reb['dnu_data']=self.log_to_lin(log_val=self.data_reb['nu_data_log'], log_err=self.data_reb['dnu_data_log'])
        
        #set original units
        for c in self.data.colnames:
            if self.data[c].unit is not None:
                self.data_reb[c] = self.data_reb[c]*self.data[c].unit

        self.data=self.data_reb
        
        self.set_fake_error(self.fake_error)

        self.data.sort('nu_data')
       
        print (section_separator)
    
    def add_systematics(self,syst,nu_range=None,data_set=None):
        """
        add systematics to errors
        
        :param syst: (float) systematic value (fractional)
        :param nu_range:  array_like of floats, [nu_min,nu_max], optional, range of frequencies to apply sistematics
        """
        
        if nu_range is not None and data_set  is not  None:
            print ("!!! error, either you provide a range of frequencies or a data_set")
            return
        
        if self.data['dnuFnu_data'] is None:
            self.data['dnuFnu_data']=np.zeros(self.data['nuFnu_data'].size)

        msk=None
        if data_set is None:
            if nu_range is None:

                msk=np.ones(len(self.data),dtype=np.bool)

            else:
                msk=get_freq_range_msk(self.data['nu_data'], nu_range)
        else:
              msk=get_data_set_msk(self.data, data_set)
        
        if msk is not None:
            self.data['dnuFnu_data'][msk]=np.sqrt(self.data['dnuFnu_data'][msk] * self.data['dnuFnu_data'][msk] + (self.data['nuFnu_data'][msk] * self.data['nuFnu_data'][msk] * syst * syst))
            self.data['nuFnu_data_log'][msk], self.data['dnuFnu_data_log'][msk]=self.lin_to_log(val=self.data['nuFnu_data'][msk], err=self.data['dnuFnu_data'][msk])

        else:
            self.data['dnuFnu_data']=np.sqrt(self.data['dnuFnu_data'] * self.data['dnuFnu_data'] + (self.data['nuFnu_data'] * self.data['nuFnu_data'] * syst * syst))
            self.data['nuFnu_data_log'], self.data['dnuFnu_data_log']=self.lin_to_log(val=self.data['nuFnu_data'], err=self.data['dnuFnu_data'])

        self.data.sort('nu_data')
        
    
    def set_error(self,error_value,nu_range=None,data_set=None,data_msk=None):
        """
         set all the paramters to same error
            
        :param error_value: float, value of the error (fractional error)
        :param nu_range:  array_like of floats, [nu_min,nu_max], optional, range of frequencies to apply the error value
        """
        #print self.data['dnuFnu_data']
        
        if nu_range is not None and data_set is not None:
            print ("!!! error, either you provide a range of frequencies or a data_set")
            return
        
        msk=None     
        if data_set is None:
            if nu_range is None:
                #for log errors
                msk=None
            else:
                msk=get_freq_range_msk(self.data['nu_data'], nu_range)
            
        else:
            msk=get_data_set_msk(self.data, data_set)
                
        if data_msk is not None:
            
            if msk is not None:
                msk=msk*data_msk
            else:
                msk=data_msk
        
        if msk is not None:
            self.data['dnuFnu_data'][msk]= error_value * self.data['nuFnu_data'][msk]
            self.data['nuFnu_data_log'][msk], self.data['dnuFnu_data_log'][msk]=self.lin_to_log(val=self.data['nuFnu_data'][msk], err=self.data['dnuFnu_data'][msk])
        
        else:
            self.data['dnuFnu_data']= error_value * self.data['nuFnu_data']
            self.data['nuFnu_data_log'], self.data['dnuFnu_data_log']=self.lin_to_log(val=self.data['nuFnu_data'], err=self.data['dnuFnu_data'])

       

    def set_fake_error(self,val):
        """
        Sets the value for the fake error
        """
        self.fake_error=val
        self.data['dnuFnu_fake_log']= np.ones(self.data['nu_data_log'].size) * self.fake_error
        self.data['dnuFnu_fake']= self.data['nuFnu_data'] * self.fake_error
        
    
    def get_data_points(self,log_log=False,skip_UL=False,frame='obs',density=False):
        """
        Gives data point
        """

        if   skip_UL==True:
            msk= self.data['UL'] == False
        else:
            msk=np.ones(self.data['nu_data_log'].size, dtype=bool)

        check_frame(frame)
        if frame == 'obs':
            if   log_log  is True:
                _x = self.data['nu_data_log'][msk]
                _dx = self.data['dnu_data_log'][msk]
                _y = self.data['nuFnu_data_log'][msk]
                _dy =  self.data['dnuFnu_data_log'][msk]

            else:
                _x = self.data['nu_data'][msk]
                _dx = self.data['dnu_data'][msk]
                _y = self.data['nuFnu_data'][msk]
                _dy = self.data['dnuFnu_data'][msk]

        elif frame == 'src':

            dl = self.cosmo.get_DL_cm(self.z)

            _x = convert_nu_to_src(self.data['nu_data'][msk], self.z, 'obs')
            _dx = convert_nu_to_src(self.data['dnu_data'][msk], self.z, 'obs')
            _y = convert_nuFnu_to_nuLnu_src(self.data['nuFnu_data'][msk], self.z, 'obs', dl)
            _dy = convert_nuFnu_to_nuLnu_src(self.data['dnuFnu_data'][msk], self.z, 'obs', dl)

            if log_log is True:
                _x, _dx = self.lin_to_log(_x, _dx)
                _y, _dy = self.lin_to_log(_y, _dy)

        else:
            unexpected_behaviour()

        if density is True:

            if log_log is True:
                _y = _y - _x
            else:
                _y = _y / _x
                _dy=_dy/_x

        return _x ,_y, _dx, _dy
    
    def show_data_sets(self):
        shown=[]
        print('current datasets')
        for entry in self.data['data_set']:
            if entry not in shown:
                shown.append(entry)
                print ('dataset', entry)
                
    
    def get_data_sets(self):
        shown=[]
        for entry in np.unique(self.data['data_set']):
            if entry not in shown:
                shown.append(entry)
        
        return shown



    def plot_sed(self,plot_obj=None,frame='obs',color=None,fmt='o',ms=4,mew=0.5,figsize=None,show_dataset=False, density=False):
        if plot_obj is None:
            plot_obj = PlotSED(frame=frame, figsize=figsize,density=density)

        if show_dataset is False:

            plot_obj.add_data_plot(self, color=color, fmt=fmt, ms=ms, mew=mew, density=density)
        else:
            for ds in self.get_data_sets():
                self.filter_data_set(filters=ds,silent=True,exclude=False)
                plot_obj.add_data_plot(self, color=color, fmt=fmt, ms=ms, mew=mew ,label='dataset %s'%ds, density=density)
                self.reset_data()
            self.reset_data()

        return plot_obj

    
    def plot_time_spans(self,save_as=None):


        fig=plt.figure(figsize=(12,9))
        ax1 = fig.add_subplot(111)
        ax1.set_xlabel('MJD')
        data_sets=self.get_data_sets()
        y=0
        line_style='-'
        for data_set in data_sets:
            print (data_set)



            T1,T2,dT,n=self.get_time_span(data_set=data_set)
            print(T1,T2,dT,n)
            if T1!=-1:
                ax1.plot([T1,T2],[y,y],label=data_set,lw=3,marker='o')
                x=(T1+T2)/2
                ax1.text(x,y+0.3,data_set+' (%d)'%n)
                y=y+1

        ax1.set_ylim(-0.5,y+0.5)

        if save_as is not None:
            fig.savefig(save_as)

        return fig
       
        
    
    def get_time_span(self,data_set=None):
        return self.find_time_span(data_set=data_set,get_values=True)
    
    def show_time_span(self,data_set=None):
        self.find_time_span(data_set=data_set,silent=False)
    
    def find_time_span(self,data_set=None,silent=True,get_values=False):
        """
        returns Tstart, Tstop, and Delta T for the full data set (if no dat_set
        is provided), or for a specific data_set
        """

        time_span_found = False

        T1 = -1
        T2 = -1
        DT = -1
        nT = 0


        if data_set is None:
            #m1= self.data['T_start'] != 0

            T1 = self.data['T_start'].min()
            T2 = self.data['T_stop'].max()
            DT = T2 - T1
            nT = self.data['T_start'].size
            time_span_found=True

            
        elif  data_set  in self.data['data_set'].astype(str):
        
            #m1= self.data['T_start'] != 0
            m2= self.data['data_set'].astype(str) == data_set
            if self.data['data_set'][m2].size>0:
                try:

                    T1=self.data['T_start'][m2].min()
                    T2=self.data['T_stop'][m2].max()
                    DT=T2-T1
                    nT=self.data['T_start'][m2].size
                    time_span_found = True
                except:
                    time_span_found = False
                    T1=-1
                    T2=-1
                    DT=-1
                    nT=0
                    print ('something wrong with T_start and T_stop columns, check the values please, for data_set=',data_set)
        else:
            time_span_found = False
            print ("no data found for this selection, data_set= ",data_set)
            if data_set not in self.data['data_set']:
                print ("the data_set %s is not present in the data"%data_set)
                print ("possible data_set: ")
                print (self.show_data_sets())
                    

        
        if silent==False and time_span_found is True:
            print ("T_start=%f T_stop=%f DT=%f, number of points=%d"%(T1, T2, DT, nT))

        if get_values==True:
            return T1,T2,DT,nT
            

    def lin_to_log(self,val=None,err=None):
        
       return lin_to_log(val=val,err=err)
       
    
    def log_to_lin(self,log_val=None,log_err=None):
        
        return log_to_lin(log_val=log_val,log_err=log_err)
        

    @property
    def gammapy_table(self):
        return self.get_gammapy_table()


    def get_gammapy_table(self):
        table = Table()
        for c in self.metadata:
            table.meta[c]=self.metadata[c]
                
        table['e_ref']=self.data['nu_data'].to("eV", equivalencies=u.spectral())
      
        table["e2dnde"] = self.data['nuFnu_data']
        table["e2dnde_err"] = self.data['dnuFnu_data']
        
        table.meta["SED_TYPE"] = "e2dnde"
        
        return table


def lin_to_log(val=None,err=None):
       
       conv_fac=np.log(10)
       
       ret_val=[]
       
       if val is not None:
           log_val=np.log10(val)
           ret_val.append(log_val)
       if err is not None:
           log_err=err/(val*conv_fac)
           ret_val.append(log_err)

       if len(ret_val)==1:
           ret_val=ret_val[0]
       
       return ret_val
   
   
def log_to_lin(log_val=None,log_err=None):
       
       conv_fac=np.log(10)
       
       ret_val=[]
       
       if log_val is not None:
           val=np.power(10.,log_val)
           
           ret_val.append(val)

       if log_err is not None:
           err=log_err*val*conv_fac
           #print err
           ret_val.append(err)

       if len(ret_val)==1:
           ret_val=ret_val[0]
       
       return ret_val


def get_freq_range_msk(x,x_range):
    msk1=x>=x_range[0]
    msk2=x<=x_range[1]
    return msk1*msk2

def get_data_set_msk(x,data_set):
    return x['data_set']==data_set